import os
import pandas as pd
import geopandas as gpd

from zipfile import ZipFile
from openpyxl import load_workbook


# HELPERS
def get_american_standers(flag):

    if flag:
        keys = {'sep': ',', 'decimal': '.'}
    else:
        keys = {'sep': ';', 'decimal': ','}
    return keys


def file_name_with_extension(path, name, ext='.csv'):
    # Reassures passed name is without extension.
    name = os.path.splitext(os.path.basename(name))[0]
    name_ext = os.path.normpath(os.path.join(path, name)) + ext
    return name_ext


# LOAD FILES
def load_text_file(path, name, cols=None, usa=False, kwargs=None):

    if kwargs is None:
        kwargs = {}

    keywords = get_american_standers(usa)
    kwargs = {**keywords, **kwargs}
    
    if 'engine' not in kwargs.keys(): 
        if 'low_memory' not in kwargs.keys():
            kwargs['low_memory'] = False
    
    # Escolhe entre ler csv ou excel.
    def read(pathname):
        if name.lower().endswith('.xlsx'):            
            kwargs.pop('low_memory', None)
            kwargs.pop('sep', None)
            kwargs.pop('decimal', None)
            df_ltf = pd.read_excel(pathname, usecols=cols, **kwargs)
        else:
            df_ltf = pd.read_csv(pathname, usecols=cols, **kwargs)
        return df_ltf
    
    # Trata caso esteja ZIPADO.
    if path[-3:] == 'zip':
        with ZipFile(os.path.normpath(path)) as z:
            with z.open(name) as myFile:
                df = read(myFile)                
    else:
        df = read(os.path.normpath(os.path.join(path, name)))

    # Drop rows that contains only NaN
    df.dropna(how="all", inplace=True)
    # Drop cols that contains only NaN
    if df.shape[0] > 1:
        df.dropna(how="all", inplace=True, axis=1)

    return df


def load_shp_file(path, name, cols=None):

    df = gpd.read_file(os.path.normpath(os.path.join(path, name)))
    try:
        # Look for the auxiliary document with the compleat name of columns.
        aux = os.path.join(path, name[:-3]+'col')
        aux = pd.read_csv(aux, header=None, engine='python')
        aux_dict = dict(zip(aux[1], aux[0]))
        df = df.rename(columns=aux_dict)
    except FileNotFoundError:
        pass
    
    if cols is not None:
        for c in cols:
            if c not in df.columns:
                msg_error = "File {} is opened somewhere or missing column {}"\
                    .format(name, c)
                raise Exception(msg_error)
            else:
                pass
        df = df.loc[:, cols]  # Return only expected columns
    return df


def load_file(path, name, expected_cols, usa, **kwargs):
    
    if name is None:
        name = os.path.basename(path)
        path = os.path.dirname(path)
    
    # Checks if name contains the suported extensions.
    extensions = ['txt', 'csv', 'shp', 'dbf', 'xlsx', 'parquet']
    try:
        ext = name.split(".")[1].lower()
        assert ext in extensions, f'{ext:} is not supported. Options are {extensions}'
    except IndexError:
        raise Exception('Param "name" was provided without extension')

    if ext in ['shp', 'dbf']:
        df = load_shp_file(path, name, expected_cols)      
        if df.geometry.isnull().all():
            del df['geometry']
    elif ext in ['parquet']:
        df = pd.read_parquet(path + '\\' + name, columns=expected_cols)
    else:        
        df = load_text_file(path, name, expected_cols, usa, **kwargs)

    return df


# SAVE FILES
def save_df_as_csv(df, path, name='test', american=False, kwargs=None):

    if kwargs is None:
        kwargs = {}

    keywords = get_american_standers(american)
    kwargs = {**keywords, **kwargs}

    # Reasures passed name is without extension
    file_name = file_name_with_extension(path, name)
    df.to_csv(file_name, **kwargs)
    return True


def save_df_as_parquet(df, path, name):
    try:
        df.columns = df.columns.astype(str)
        file_name = file_name_with_extension(path, name, ext='.parquet')
        df.to_parquet(file_name, compression=None)
    except ValueError:
        print('Error while creating a usable Parquet, creating a CSV instead...')
        save_df_as_csv(df, path, name, american=False)
        
    return True


def save_df_as_shp(df, path, name):
    # There is a limitation with SHPs that only saves columns with 10charcMax.
    # So we are doing a gambiarra...
    
    col_names = list(df.columns)
    col_names.remove('geometry')
    
    aux_dict = dict(zip(col_names, ['c'+str(x) for x in list(range(0, len(col_names)))]))
    temp = pd.Series(aux_dict).to_frame()
    temp.to_csv(os.path.normpath(os.path.join(path, name + '.col')), 
                header=False, index=True)
    
    df = df.rename(columns=aux_dict)    
    df.to_file(os.path.normpath(os.path.join(path, name + '.shp')), 
               driver='ESRI Shapefile')
    return True


def save_df_as_excel(df, path, name, sheet_name='Sheet1', startrow=None,
                     truncate_sheet=False, **to_excel_kwargs):

    filename = file_name_with_extension(path, name, ext='.xlsx')
           
    # Ignore [engine] parameter if it was passed.
    if 'engine' in to_excel_kwargs:
        to_excel_kwargs.pop('engine')

    try:
        # Try to open an existing workbook.
        load_workbook(filename)
        writer = pd.ExcelWriter(filename, engine='openpyxl')

        # Get the last row in the existing Excel sheet.
        # If it was not specified explicitly.
        if startrow is None and sheet_name in writer.book.sheetnames:
            startrow = writer.book[sheet_name].max_row

        # Truncate sheet.
        if truncate_sheet and sheet_name in writer.book.sheetnames:
            # Index of [sheet_name] sheet.
            idx = writer.book.sheetnames.index(sheet_name)
            # Remove [sheet_name].
            writer.book.remove(writer.book.worksheets[idx])
            # Create an empty sheet [sheet_name] using old index.
            writer.book.create_sheet(sheet_name, idx)

        # Copy existing sheets.
        writer.sheets = {ws.title: ws for ws in writer.book.worksheets}
    except FileNotFoundError:
        # File does not exist yet, we will create it.
        pass

    if startrow is None:
        startrow = 0

    writer = pd.ExcelWriter(filename, engine='openpyxl')

    # Write out the new sheet.
    df.to_excel(writer, sheet_name, startrow=startrow, 
                index=False, **to_excel_kwargs)

    # Save the workbook.
    writer.save()
    return None


def save_zip(df, path, zip_file, file_name):
    # First needs to save the file normally.
    temp_file = file_name_with_extension(path, file_name, ext='.txt')
    df.to_csv(temp_file, index=None, sep=';', mode='w', float_format='%.6f',
              decimal=',', date_format='%d/%m/%Y %H:%M:%S')
    # Then add its to the zip.
    zip_path = file_name_with_extension(path, zip_file, ext='.zip')
    with ZipFile(zip_path, 'a') as z:
        z.write(temp_file, file_name + '.txt')
    # Now delete the temp file.
    os.remove(temp_file)

    return True



from .utils import charts as ch
from .utils import maps as mp
import os
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib_scalebar.scalebar import ScaleBar
import numpy as np
import geopandas as gpd
import warnings

warnings.simplefilter(action='ignore', category=FutureWarning)


def hist(what, bins=5, lwl=float('-inf'), upl=float('inf'),
         legenda=None, xlabel=None, subtitle='', title='', report_nan=True,
         comentario='', source='', legend_control='%V'):

    """
    Plot a cute histogram based on the source data the user provide.

        Parameters
        ----------
        what: DataFrame or Series
            The data that the user want to be the histogram base.
        bins: integer
            Integer number that represent the number of requisited
            bins that the user want to be in the histogram.
        lwl: float
            Float number to represent the lower limit.
        upl: float
            Float number to represent the upper limit.
        legenda: str
            String value that represents the title of the plot
            legend.
        xlabel: str
            String value responsible for the xlabel title.
        subtitle: str
            String value responsible for the graph subtitle.
        title: str
            String value responsible for the graph title.
        report_nan: bool
            Boolean value responsible to represent if the user
            want or not the column reporting the number of NaN
            values.
        comentario: str
            String value responsible to plot a comment it the bottom
            of the graph.
        source: str
            String value that indicates the source of the data.
        legend_control: str
            String value that indicates the type of graph axis and
            labels displayed.
            V - For just values.
            % - For just percentages.
            V% - For values as labels and percentages as axis.
            %V - For percentages as labels and values as axis.

        Returns
        -------

        fig : Matplotlib figure type.
            Figure containing the results os the histogram plot.

    """

    source = source

    what = ch.check_what(what)
    what_ticks = what.copy()

    fig, ax1, ax2 = ch.create_figure(str(title), str(subtitle), report_nan, source=source, comentario=comentario)

    cols_ref = list(what.columns)
    what, flag_lw, flag_up = ch.clip2plot(what, ch.make_bins(bins, what), lwl, upl)
    bins = ch.make_bins(bins, what)

    stat = ch.get_null_stat(what, legenda)
    what = ch.get_what2plot(what, bins)
    what, legenda = ch.redef_cols(what, cols_ref, legenda)

    if report_nan:
        per = True
        stat.plot.bar(stacked=True, ax=ax2, legend=True,
                      color=['#1C1C1C', '#E6E6E6'])
        ax2.set_xticklabels(stat.index, rotation=0)
        ax2.legend(loc='lower center')
        ch.hide_grid(ax2)
        ch.bars_auto_label(ax2, what_ticks, per, decimal=0, pos_above=True)

    if legend_control == '%V':

        per = True

        ax1.patch.set_facecolor('none')
        if ax2 is not None:
            ax2.patch.set_facecolor('none')
        ax_aux = ax1.twinx()

        what.plot.bar(stacked=True, ax=ax_aux, legend=legenda)
        if xlabel is not None:
            ax1.set_xlabel(xlabel, fontweight='bold', fontsize=14, color='#434853')
        ch.hide_grid(ax1)
        ch.bars_auto_label(ax_aux, what_ticks, per, decimal=1, pos_above=not legenda)

        ax_aux.set_ylabel('N° Ocorrências', rotation=0)
        ax_aux.yaxis.set_label_coords(1, 1.06)
        ax_aux.patch.set_facecolor('none')
        ticks = list(ax_aux.get_yticks())
        new_ticks = [tick * what_ticks.shape[0] for tick in ticks]
        ax_ticks = ax1.twinx()
        ax_ticks.set_yticks(new_ticks)

        ax_ticks.grid(color='grey', linestyle='-', linewidth=0.5)
        ax_ticks.set_zorder(ax_ticks.get_zorder() - 1)
        ax_ticks.set_yticklabels(['{:,.0f}'.format(tick) for tick in new_ticks], color='grey')
        ch.hide_grid(ax_aux)

    elif legend_control == '%':

        per = True

        what.plot.bar(stacked=True, ax=ax1, legend=legenda)
        if xlabel is not None:
            ax1.set_xlabel(xlabel, fontweight='bold', fontsize=14, color='#434853')
        ch.hide_grid(ax1)
        ch.bars_auto_label(ax1, what_ticks, per, decimal=1, pos_above=not legenda)

    elif legend_control == 'V':
        per = False
        what.plot.bar(stacked=True, ax=ax1, legend=legenda)
        if xlabel is not None:
            ax1.set_xlabel(xlabel, fontweight='bold', fontsize=14, color='#434853')
        ch.hide_grid(ax1)
        ch.bars_auto_label(ax1, what_ticks, per, decimal=1, pos_above=not legenda)

    elif legend_control == 'V%':

        per = False

        ax1.patch.set_facecolor('none')
        if ax2 is not None:
            ax2.patch.set_facecolor('none')
        ax_aux = ax1.twinx()

        what.plot.bar(stacked=True, ax=ax_aux, legend=legenda)
        if xlabel is not None:
            ax_aux.set_xlabel(xlabel, fontweight='bold', fontsize=14, color='#434853')
        ch.hide_grid(ax1)
        ch.bars_auto_label(ax_aux, what_ticks, per, decimal=1, pos_above=not legenda)

        ax_aux.set_ylabel('N° Ocorrências', rotation=0)
        ax_aux.yaxis.set_label_coords(1, 1.06)
        ax_aux.patch.set_facecolor('none')

        ax_aux.grid(color='grey', linestyle='-', linewidth=0.5)
        ax_aux.set_zorder(ax_aux.get_zorder() - 1)
        ax_aux.set_yticklabels(['{:.2f}%'.format(tick * 100) for tick in list(ax_aux.get_yticks())], color='grey')

    if bins[1] - bins[0] == 1:
        ch.adjust_xlabel(ax1, bins[:-1], flag_lw, flag_up)
    else:
        ch.adjust_xlabel(ax1, bins, flag_lw, flag_up)
    
    return fig


def mapa(shapes, path=None, title='', subtitle='', col_lin=None, dir2dashed=False, col_pts=None,
         col_size=3, coords=None, dir_col='DIR', col_zns=None, heat=None):

    """
    Function used to plot maps based on multiple inputed GeoDataFrames.

        Parameters
        ----------
        shapes: list
            List of the GeoDataFrames that the user wants to plot.
        path: str
            String value with the file output path for the maps
            ploted using this function.
        title: str
            Title of the map.
        subtitle: str
            Subtitle of the map.
        col_lin: str
            String value responsible for doing two different things:
            1 - Determinate what color you want the linestrings to be.
            or
            2 - Determinate what GeoDataFrame column you want to be used
        dir2dashed: bool
            Boolean value to indicate if the analized linestrings have
            multiple directions.
        col_pts: str
            String value responsible for doing two different things:
            1 - Determinate what color you want the points to be.
            or
            2 - Determinate what GeoDataFrame column you want to be used
                as categoric values for the points.
        col_size: int or str
            Determinate the size of the point ploted values.
        coords: list
            List of GeoDataFrame columns responsible to share the coordinates
            of the point data that are going to be ploted if there wasn't any
            geometry column.
        dir_col: str
            String value with the name of the column that have the direction
            data.
        col_zns: str
            String value responsible for doing two different things:
            1 - Determinate what color you want the polygons to be.
            or
            2 - Determinate what GeoDataFrame column you want to be used
                as categoric values for the polygons.
        heat: str
            String value with the seaborn color pallete used for
            cloropletic maps. It will create a cloropletic zone
            classification

        Returns
        -------

        fig : Matplotlib figure type.
            Figure containing the ploted map.

    """

    lines_plot = gpd.GeoDataFrame(columns=['geometry'])
    points_plot = gpd.GeoDataFrame(columns=['geometry'])
    zones_plot = gpd.GeoDataFrame(columns=['geometry'])
    lines = gpd.GeoDataFrame(columns=['geometry'])
    points = gpd.GeoDataFrame(columns=['geometry'])
    zones = gpd.GeoDataFrame(columns=['geometry'])

    for shape in shapes:

        if 'geometry' not in shape.columns:
            points_gdf = mp.config_crs(shape, coords)
            points_gdf = points_gdf.set_crs('epsg:4674', allow_override=True)
            points_plot = points_plot.append(points_gdf, ignore_index=True)
            points_plot = gpd.GeoDataFrame(points_plot['geometry'])
            shape_type = points_plot['geometry'].geom_type.to_list()
            shape_type = list(set(shape_type))
            shape = points_gdf.copy()

        if 'geometry' in shape.columns:
            shape_type = shape['geometry'].geom_type.to_list()
            shape_type = list(set(shape_type))

            if len(shape_type) == 1:
                if shape_type[0] == 'Point' or shape_type[0] == 'MultiPoint':
                    points_gdf = shape
                    points_gdf = points_gdf.set_crs('epsg:4674', allow_override=True)
                    points_plot = points.append(points_gdf, ignore_index=True)
                    points_plot = gpd.GeoDataFrame(points_plot['geometry'])
                    points = points.append(points_plot)
                elif shape_type[0] == 'Polygon' or shape_type[0] == 'MultiPolygon':
                    zones_gdf = shape
                    zones_gdf = zones_gdf.set_crs('epsg:4674', allow_override=True)
                    zones_plot = zones_plot.append(zones_gdf, ignore_index=True)
                    zones_plot = gpd.GeoDataFrame(zones_plot['geometry'])
                    zones = zones.append(zones_plot)
                elif shape_type[0] == 'LineString' or shape_type[0] == 'MultiLineString':
                    lines_gdf = shape
                    lines_gdf = lines_gdf.set_crs('epsg:4674', allow_override=True)
                    lines_plot = lines_plot.append(lines_gdf, ignore_index=True)
                    lines_plot = gpd.GeoDataFrame(lines_plot['geometry'])
                    lines = lines.append(lines_plot)

            elif len(shape_type) == 2:
                if shape_type[0] == 'Point' and shape_type[0] == 'MultiPoint':
                    points_gdf = shape
                    points_gdf = points_gdf.set_crs('epsg:4674', allow_override=True)
                    points_plot = points.append(points_gdf, ignore_index=True)
                    points_plot = gpd.GeoDataFrame(points_plot['geometry'])
                    points = points.append(points_plot)
                elif shape_type[0] == 'Polygon' and shape_type[0] == 'MultiPolygon':
                    zones_gdf = shape
                    zones_gdf = zones_gdf.set_crs('epsg:4674', allow_override=True)
                    zones_plot = zones_plot.append(zones_gdf, ignore_index=True)
                    zones_plot = gpd.GeoDataFrame(zones_plot['geometry'])
                    zones = zones.append(zones_plot)
                elif shape_type[0] == 'LineString' and shape_type[0] == 'MultiLineString':
                    lines_gdf = shape
                    lines_gdf = lines_gdf.set_crs('epsg:4674', allow_override=True)
                    lines_plot = lines_plot.append(lines_gdf, ignore_index=True)
                    lines_plot = gpd.GeoDataFrame(lines_plot['geometry'])
                    lines = lines.append(lines_plot)

            elif len(shape_type) > 2:
                lines_gdf = shape[(shape['geometry'].geom_type ==
                                   'LineString') | (shape['geometry'] == 'MultiLineString')].copy()
                points_gdf = shape[(shape['geometry'].geom_type == 'Point') |
                                   (shape['geometry'] == 'MultiPoint')].copy()
                zones_gdf = shape[(shape['geometry'].geom_type == 'Polygon') |
                                  (shape['geometry'] == 'MultiPolygon')].copy()
                points_gdf = points_gdf.set_crs('epsg:4674', allow_override=True)
                zones_gdf = zones_gdf.set_crs('epsg:4674', allow_override=True)
                lines_gdf = lines_gdf.set_crs('epsg:4674', allow_override=True)
                points_plot = points.append(points_gdf, ignore_index=True)
                points_plot = gpd.GeoDataFrame(points_plot['geometry'])
                lines_plot = lines_plot.append(lines_gdf, ignore_index=True)
                lines_plot = gpd.GeoDataFrame(lines_plot['geometry'])
                zones_plot = zones_plot.append(zones_gdf, ignore_index=True)
                zones_plot = gpd.GeoDataFrame(zones_plot['geometry'])
                zones = zones.append(zones_plot)
                lines = lines.append(lines_plot)
                points = points.append(points_plot)

        if dir2dashed and dir_col in shape.columns:
            lines_plot = lines_plot.merge(shape[['geometry', dir_col]], how='right', left_on='geometry',
                                          right_on='geometry')
            lines = lines.append(lines_plot)
        if col_pts in shape.columns and (shape_type[0] == 'Point' or shape_type[0] == 'MultiPoint'):
            points_plot = points_plot.merge(shape[['geometry', col_pts]], how='right', left_on='geometry',
                                            right_on='geometry')
            points = points.append(points_plot)
        if col_lin in shape.columns and (shape_type[0] == 'LineString' or shape_type[0] == 'MultiLineString'):
            lines_plot = lines_plot.merge(shape[['geometry', col_lin]], how='right', left_on='geometry',
                                          right_on='geometry')
            lines = lines.append(lines_plot)
        if col_zns in shape.columns and (shape_type[0] == 'Polygon' or shape_type[0] == 'MultiPolygon'):
            zones_plot = zones_plot.merge(shape[['geometry', col_zns]], how='right', left_on='geometry',
                                          right_on='geometry')
            zones = zones.append(zones_plot)
        if col_size in shape.columns and (shape_type[0] == 'Point' or shape_type[0] == 'MultiPoint'):
            points_plot = points_plot.merge(shape[['geometry', col_size]], how='right', left_on='geometry',
                                            right_on='geometry')
            points = points.append(points_plot)

    points.dropna(inplace=True)
    lines.dropna(inplace=True)
    zones.dropna(inplace=True)

    fig, ax = plt.subplots(figsize=(8, 8), dpi=150)
    fig.suptitle(f'{title}\n{subtitle}', fontsize=20, ha='left', va='center', x=0.125, y=0.95, **{'fontname': 'Lato'})
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    ax.add_artist(ScaleBar(1))

    if lines is not None and len(lines) > 0:
        lines, col_lin, color = mp.group_data(lines, col_lin, 'Rotas')
        if dir2dashed and dir_col in lines.columns:
            lines1 = lines.loc[lines[dir_col] == 1, :]
            lines2 = lines.loc[lines[dir_col] == 2, :]
            lines2plot = [lines1, lines2]
        else:
            lines2plot = [lines]

        for i, geo in enumerate(lines2plot):
            color_attrs = mp.get_colors(geo, col_lin)
            for ctype, data in geo.groupby(col_lin):
                if color is None:
                    data.plot(color=color_attrs[ctype],
                              label=ctype,
                              ax=ax,
                              linewidth=1, linestyle='-' * (i + 1),
                              aspect='equal')
                else:
                    data.plot(color=color,
                              label=ctype,
                              ax=ax,
                              linewidth=1, linestyle='-' * (i + 1),
                              aspect='equal')

    plt.setp(ax.get_xticklabels(), visible=False)
    plt.setp(ax.get_yticklabels(), visible=False)
    x, y, arrow_length = 0.1, 0.15, 0.1
    ax.annotate('N', xy=(x, y), xytext=(x, y - arrow_length),
                arrowprops=dict(facecolor='black', width=5, headwidth=15),
                ha='center', va='center', fontsize=20,
                xycoords=ax.transAxes)

    def corresp(var):
        if var == 1:
            return 3
        if var == 2:
            return 20
        if var == 3:
            return 40
        if var == 4:
            return 100
        else:
            return 0

    if points is not None and len(points) > 0:
        points, col_pts, color = mp.group_data(points, col_pts, 'POIs')
        color_attrs = mp.get_colors(points, col_pts)

        labels = [0, 200, 400, 800]
        labels = [round(x) for x in labels]

        if isinstance(col_size, str):
            points['size'] = np.digitize(points[col_size], labels)
            labels = [' : Menos que 200', ' : Menos que 400', ' : Menos que 800', ' : Mais que 800']

        for ctype, data in points.groupby(col_pts):
            if isinstance(col_size, str):
                for size, data2 in data.groupby('size'):
                    if color is None:
                        data2.plot(color=color_attrs[ctype],
                                   label=ctype + labels[size - 1],
                                   ax=ax,
                                   markersize=corresp(size),
                                   aspect='equal')
                    else:
                        data2.plot(color=color,
                                   label=ctype + labels[size - 1],
                                   ax=ax,
                                   markersize=corresp(size),
                                   aspect='equal')
            else:
                if color is None:
                    data.plot(color=color_attrs[ctype],
                              label=ctype,
                              ax=ax, zorder=10,
                              markersize=col_size,
                              aspect='equal')
                else:
                    data.plot(color=color,
                              label=ctype,
                              ax=ax, zorder=10,
                              markersize=col_size,
                              aspect='equal')

    if zones is not None and len(zones) > 0:
        zones, col_zns, color = mp.group_data(zones, col_zns, 'Zonas')
        color_attrs = mp.get_colors(zones, col_zns, heat_def=heat)
        for ctype, data in zones.groupby(col_zns):
            if color is None:
                data.plot(color=color_attrs[ctype],
                          label=ctype,
                          ax=ax,
                          alpha=0.3, edgecolor='black',
                          aspect='equal')
            else:
                data.plot(color=color,
                          label=ctype,
                          ax=ax,
                          alpha=0.3, edgecolor='black',
                          aspect='equal')

    ax.legend(fontsize=6, loc='upper left')  # attention
    plt.axis('equal')
    if path is not None:
        file_name = os.path.join(path, 'chart_' + title + '.png')
        plt.savefig(file_name, dpi=300, bbox_inches='tight',
                    pad_inches=0.5, facecolor='#F2F2F2')

    return fig


def plot_sidemap(map_left, map_right, file_name=None):

    """
    Function that is responsible for plot two early created maps side by side.

        Parameters
        ----------
        map_left: Matplotlib figure type
            Figure of the map you want to be plotted on the left
            side of the new figure.
        map_right: Matplotlib figure type
            Figure of the map you want to be plotted on the right
            side of the new figure.
        file_name: str
            String value with the path and file name of the exported
            new figure.

        Returns
        -------

        fig : Matplotlib figure type.
            Figure containing the two ploted maps, side by side.

    """

    dpi = 300

    backend = mpl.get_backend()
    mpl.use('agg')

    c1 = map_left.canvas
    c2 = map_right.canvas
    c1.draw()
    c2.draw()

    a1 = np.array(c1.buffer_rgba())
    a2 = np.array(c2.buffer_rgba())
    a = np.hstack((a1, a2))

    mpl.use(backend)
    fig, ax = plt.subplots(figsize=(6000 / dpi, 3000 / dpi), dpi=dpi)
    fig.subplots_adjust(0, 0, 1, 1)
    ax.set_axis_off()
    ax.matshow(a)
    if file_name is not None:
        plt.savefig(file_name, dpi=300, bbox_inches='tight',
                    pad_inches=0.7, facecolor='#F2F2F2')

    return fig



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 12 10:32:35 2022

@author: bcalazans

Módulo da regressão linear
> útil para a determinação de GERAÇÂO DE VIAGENS no modelo 4 etapas

Dado um conjunto de variáveis socioeconomicas e atração/produção das zonas,
gera todas as regressões estatisticamente plausíveis e solta uma tabela para
avalição crítica dos resultados

É possível gerar relatório em HTML de regress~eos específicas, para visualizar 
os testes estatísticos realizados no processo
"""

import os
import numpy as np
import pandas as pd
import statsmodels.api as sm
import matplotlib.pyplot as plt
import seaborn as sns
import warnings
from tqdm import tqdm

# Systool modules
from .utils import linear_regression_make as lrm
from .utils.linear_regression_plot import plot_regression, give_name
from .utils import report


def fit_model(x, y, intercept=False, cov_type='nonrobust', plot=False, path=None):
    """
    Executa regressão e testa retirada de outliers
    Retorna dois modelos, um completo e outro sem outliers

    Parameters
    ----------
    x : pd.Series ou pd.DataFrame
        Variáveis independentes do modelo de regressão.
    y : pd.Series
        Variável dependente do modelo de regressão. Geralmente, ATRAÇÃO ou PRODUÇÃO de viagens.
    intercept : bool, optional
        True para forçar que regressão cruze o eixo 0. The default is False.
    cov_type : str, optional
        See regression.linear_model.RegressionResults for a description of the available covariance estimators...
        The default is 'nonrobust'.
    plot : bool, optional
        Gera um arquivo HTML com vários testes estatísticos. The default is False.
    path : str, optional
        Salva o arquivo HTML no caminho indicado. The default is None (does not save file)

    Returns
    -------
    model : container or StatsModelResults
        Modelo com todos os registors.
    model_out : container or StatsModelResults
        Modelo otimizado sem outliers.

    """
    # TODO - mudar o default para o default so stats

    if isinstance(x, pd.Series):
        x = pd.DataFrame(x)
    # if isinstance(y,pd.Series): y = pd.DataFrame(y)
    # http s://stats.stackexchange.com/questions/47913/pandas-statsmodel-scikit-learn
    y = y.fillna(0)
    
    # TODO -- if X has N/a drop zones
    # TODO -- if ZN missing in X or Y, drop in both    
    
    # substitui zeros com pequeno ruído para evitar SingularMAtrix error
    x = x+0.00001*np.random.rand(*x.shape)
    y = y+0.00001*np.random.rand(*y.shape)
    
    if intercept:
        model = sm.OLS(y, sm.add_constant(x)).fit(cov_type=cov_type)
    else:
        model = sm.OLS(y, x).fit(cov_type=cov_type)
    
    model_out, dfr, xx, yy, grafico_html = lrm.find_outliers(model, x, y, intercept, cov_type, plot)
    if plot: 
        # Monta o nome do HTML e gráficos
        if path is None:
            path = os.getcwd()
        name = give_name(x, y, intercept)
        
        # plota gráficos e salva HTML
        list_2_plots_a = plot_regression(model, x, y, intercept)
        
        if model_out is None:
            report.save_html(path, name, list_2_plots_a)
        else:
            list_2_plots_b = plot_regression(model_out, xx, yy, intercept)
            dict_pages = {'Modelo completo': list_2_plots_a,
                          'Modelo sem outliers': list_2_plots_b,
                          'Grafico R²': grafico_html}
            report.save_html(path, name, dict_pages)    
        
        # tenta abrir HTML automaticamente
        report.open_html(path, name)
    return model, model_out


def loop_models(df, xcols, ycol, mask=None, keep_all=False, cut_r=0.5, force_intercept=False, kargs={}):
    """
    Realiza a regressão de todas as combinações não correlacionadas
    Faz testes para ver se a regressão é válida ou não e salva informações em
    um dataframe para investigação

    Parameters
    ----------
    df : TYPE
        DESCRIPTION.
    mask : boolean series, OPTIONAL
        Pass a mask to dataframe and rows can be ignored
    xcols : list
        List of column names for independent stuff.
    ycol : string
        NAme of column for the dependent variable.
    keep_all : TYPE, optional
        Manter todas as regressões, ainda que não tenham passado nos testes. 
        The default is False.
    cut_r : TYPE, optional
        Rsquared de linha de corte. Valores maiores indicam que regressão pode
        melhorar tirando outliers. Realiza então a regressão do outlier. 
        The default is 0.5.
    force_intercept : bool
        if TRUE best model is alwys the model that intercepts
    kargs : DICT, optional
        Parâmetros para fit_model: 
            > intercept (True or Flse)
            > plot (True or False)
            > covType (nonrobust, HC3, etc)

    Returns
    -------
    df_regs : pandas dataFrame
        Contains all regressions found with some indicator to help us choose for the best one.

    """
    warnings.filterwarnings('ignore')
    if mask is None:
        mask = np.ones(len(df), dtype=bool)
    # TODO -- criar status se reg teve outliers eliminados ou não
    print('Default regressions performed with {:.1%} of data'.format(mask.sum()/len(mask)))
    df_regs = pd.DataFrame(columns=['x', 'Rsquared', 'intercept', 'covType', 'ZONAS',
                                    'PASS_TESTS', 'NUM_COEF_NEGATIVO'])
    # calcula todas possibilidades de regressões em que não há muita correlação
    comb = lrm.get_combinations(xcols)
    corr_pairs = lrm.get_correlated_pairs(df, xcols)
    comb = lrm.drop_combinations(comb, corr_pairs, xcols)

    # calcula um modelo para cada possibilidades
    for c in tqdm(comb, position=0, leave=True, desc='Making regressions...'):
        
        if isinstance(c, tuple):
            c = list(c)
        else:
            c = [c]  # need this for one column value
        
        x = df.loc[mask, c]
        y = df.loc[mask, ycol]
       
        # make two possible regressions
        model_a, model_a_out = fit_model(x, y, intercept=True, **kargs)
        model_b, model_b_out = fit_model(x, y, intercept=False, **kargs)
           
        # choose best model
        if force_intercept:
            model = model_a
            model_out = model_a_out
            intercept = True
        else:
            if model_a.rsquared > model_b.rsquared:
                model = model_a
                model_out = model_a_out
                intercept = True
            else: 
                model = model_b
                model_out = model_b_out
                intercept = False
        
        # se modelo é aceitavel, salva as regressões
        if model.rsquared > cut_r:
            df_regs = lrm.save_model(model, df_regs, intercept, lrm.test_conditions(model))
            df_regs = lrm.save_model(model_out, df_regs, intercept, lrm.test_conditions(model_out))

    print('{:5} regressões decentes (R² > {:}) foram geradas'.format(len(df_regs), cut_r))
    print('{:5} regressões passaram nos testes'.format((df_regs.PASS_TESTS == 'SIM').sum()))
    # TODO - imprimir aqui as condições impostas para passar nos testes
    
    if not keep_all:
        df_regs = df_regs[df_regs.PASS_TESTS == 'SIM']
    
    df_regs['numXvars'] = df_regs.x.str.len().astype(int)
    df_regs['numZones'] = df_regs.ZONAS.str.len()
    df_regs['Rsquared'] = df_regs['Rsquared'].astype('float').round(4)
    df_regs.sort_values('Rsquared', ascending=False, inplace=True)
    
    if len(df_regs) > 0:
        # plot the R results  
        plt.subplots()
        sns.set_style("ticks")
        sns.swarmplot(data=df_regs, x="numXvars", y="Rsquared", 
                      hue="PASS_TESTS",
                      palette={'NOP': 'red', 'SIM': 'green'}
                      )
        sns.despine(trim=True)
        
        # plot the usage of the variabels
        # mostra quantas vezes uma variável X foi usada em um modelo contendo numXvars
        usage = df_regs.set_index('numXvars').x.explode().reset_index().groupby(['numXvars', 'x']).size().unstack().T
        plt.subplots()
        sns.heatmap(usage, cmap='Blues', annot=True)
    return df_regs



import geopandas as gpd
from shapely.geometry import Point
import seaborn as sns


def config_crs(geo, aux=None):
    if not (isinstance(geo, gpd.GeoDataFrame)):
        geometry = [Point(xy) for xy in zip(geo[aux[1]], geo[aux[0]])]
        geo = geo.drop(aux, axis=1)
        geo = gpd.GeoDataFrame(geo, geometry=geometry)
    return geo


def get_main_frame(frames):
    # an order to have a main frame
    for f in frames:
        if f is not None and len(f) > 0:
            return f


def group_data(geo, col, label):
    if col is None:
        # creates a temp col with a unique atribute
        color = None
        col = 'temp'
        geo[col] = label
    elif col not in geo.columns:
        color = col
        col = 'temp'
        geo[col] = label
    else:
        color = None
    return geo, col, color


def get_colors(geo, col, heat_def=None):

    if heat_def is None:

        urban_systra = ["#C3423F", "#9BC53D", "#fdbf11", "#FDE74C", "#5BC0EB", "#404E4D"]
        keys = list(geo[col].unique())
        values = sns.color_palette(urban_systra, n_colors=len(keys))
        # values = sns.color_palette("Paired", n_colors=len(keys))
        color_attrs = dict(zip(keys, values))

    if heat_def is not None:

        keys = list(sorted(set(geo[col].unique())))
        values = sns.color_palette(heat_def, n_colors=len(keys))
        # values = sns.color_palette("Paired", n_colors=len(keys))
        color_attrs = dict(zip(keys, values))

    return color_attrs



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 12 10:28:14 2022

@author: bcalazans & MoleDownTheHole (pcardoso)

Módulo com funções úteis no tratamento de dados tabulados.

"""
# Python Modules
import os
import geopandas as gpd
import pandas as pd
import numpy as np
from pyproj import Proj
from shapely.geometry import LineString, Point

# Systool Modules
from .utils import flatt_geom as fg


def get_line_zoning_and_muni(line_shp, zone_shp, column_id_zone='ID_ZONE', column_nm_muni='NM_MUNI'):

    """
    Tag all the Zones ID and Municipalities to each line that are intersected by it.

    Parameters
    ----------
    line_shp: GeoDataFrame
              GeoDataFrame containing the LineStrings.
    zone_shp: GeoDataFrame
              GeoDataFrame containing the Zones and all municipalities.
    column_id_zone: String
                    Name of the column containing the zone ID.
    column_nm_muni: String
                    Name of the column containing the zone municipalities.

    Returns
    -------

    gdf: GeoDataFrame
         GeoDataFrame with all zone intersection occurrence in each line,
         with municipalities as well.

    """

    gdf = gpd.sjoin(line_shp, zone_shp[[column_id_zone, column_nm_muni, 'geometry']], how='left')

    return gdf


def calc_dist_euclidean(df, x1, x2, y1, y2):

    """
    Calculate the euclidean distances between given coordinates of a Dataframe.

    Parameters
    ----------

    df: DataFrame
        DataFrame containing the coordinates.
    x1: String
        Dataframe column name with the
        origin Longitudes.
    x2: String
        Dataframe column name with the
        destination Longitudes.
    y1: String
        Dataframe column name with the
        origin Latitudes.
    y2: String
        Dataframe column name with the
        destination Latitudes.

    Returns
    -------

    result: Float
            Result of the euclidean distance.

    """

    df[x1] = pd.to_numeric(df[x1], errors='coerce')
    df[x2] = pd.to_numeric(df[x2], errors='coerce')
    df[y1] = pd.to_numeric(df[y1], errors='coerce')
    df[y2] = pd.to_numeric(df[y2], errors='coerce')

    result = np.sqrt((df[x1] - df[x2]) ** 2 + (df[y1] - df[y2]) ** 2)

    return result


def calc_dist_milepost(gdf, mask=None, x='UTMx', y='UTMy', new_col='dist'):

    """

    Calculates the distance in meters of a point along the line (milepost).
    ATTENTION! GeoDataFrame must be projected in mercator.

    Parameters
    ----------
    gdf: GeoDataFrame
         GeoDataFrame containing the geographical data
         that the milepost are going to be calculated.
    mask: Pandas Bool Series
          Pandas Bool Series to mask the given
          GeoDataFrame.
    x: String
       Longitude column name.
    y: LineString
       Latitude column name.
    new_col: String
             Column name for the calculated milepost.

    Returns
    -------

    gdf: GeoDataFrame
         Inputted GeoDataFrame now with the mileposts
         distance column.

    """

    if mask is None:
        mask = np.ones(len(gdf), dtype=bool)

    # uses pe as a auxiliary column
    gdf['pe'] = [Point(xx, yy) for xx, yy in zip(gdf[x], gdf[y])]

    # calculate the distance as a new column
    gdf.loc[mask, new_col] = (gpd.GeoSeries(gdf.loc[mask, 'geometry']).project(gpd.GeoSeries(gdf.loc[mask, 'pe'])))
    gdf.loc[~mask, new_col] = None  # garantees if colum already existed to erase

    gdf = gdf.drop('pe', 1)

    return gdf


def calc_dist_orthogonal(df, mask=None, x='UTMx', y='UTMy', new_col='dist'):

    """
    Calculates the minimum distance between a coord (x,y) and a line 
    (i.e., the orthogonal projection).
    ATTENTION ! geo.DataFrame must be projected in mercator.

    Parameters
    ----------
    df: GeoDataFrame
        GeoDataFrame containing the geometry
        lines and all coordinates.
    mask: Pandas Bool Series
          Pandas Bool Series to mask the given
          GeoDataFrame.
    x: String
       Longitude column name.
    y: LineString
       Latitude column name.
    new_col: String
             Column name for the calculated milepost.

    Returns
    -------

    gdf: GeoDataFrame
         Inputted GeoDataFrame now with the mileposts
         distance column.

    """

    # TODO check the CRS of the dataframe using geopandas to see if it is in UTM
    # if not, reproject it for UTM and TELL THE USER that the answer was reprojected

    if mask is None:
        mask = np.ones(len(df), dtype=bool)

    # uses pe as a auxiliary column

    df['pe'] = [Point(xx, yy) for xx, yy in zip(df[x], df[y])]

    # calculate the distance as a new columns

    df.loc[mask, new_col] = (gpd.GeoSeries(df.loc[mask, 'geometry']).distance(gpd.GeoSeries(df.loc[mask, 'pe'])))

    gdf = df.drop('pe', 1)

    return gdf


# TODO - entender prq isso não está com a atualização do geopandas
# ajustar para geopandas novo q pedro já fez no MovePass

def convert2utm(path, file, utm_zone=23, set_south=True):

    """
    Converts a given shapefile to UTM.
    Saves the file converted in the same path with utm_ as prefix name.

    Parameters
    ----------
    path: String
          Path containing the file to be open
          without the file name and extension.
    file: String
          File name with extension.
    utm_zone: Int
              UTM Zone code.
    set_south: Bool
               True if located in southern hemisphere,
               False if located in northern hemisphere.

    """

    shp = gpd.read_file(os.path.join(path, file))
    my_string = "+proj=utm +zone=" + str(utm_zone) + " +ellps=WGS84 +datum=WGS84 +units=m +no_defs"

    if set_south:

        my_string = my_string + ' +south'

    if shp.crs is None:
        shp.crs = "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs"

    shp = shp.to_crs(my_string)
    shp.to_file(driver='ESRI Shapefile',
                filename=os.path.join(path, 'utm_' + file))
    return


def convert2degree(path, file):

    """
    Converts a given shapefile to degrees, i.e. Lat and Long.
    Saves the file converted in the same path with degree_ as prefix name.

    Parameters
    ----------
    path: String
          Path containing the file to be open
          without the file name and extension.
    file: String
          File name with extension.

    """

    shp = gpd.read_file(os.path.join(path, file))
    my_string = "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs"

    shp.to_crs(my_string)
    shp.to_file(driver='ESRI Shapefile',
                filename=os.path.join(path, 'degree_' + file))

    return


def convert_dataframe(df, convert_to, utm=['UTMx', 'UTMy'], lat_lon=['LAT', 'LON'],
                     utm_zone=23, set_south=True):
    """
    Convert columns of DataFrame from/to degrees/UTM.

    Parameters
    ----------
    df: DataFrame
        DataFrame containing the columns
        that are going to be converted.
    convert_to: String
                Text value with the convertion type.
                Must be 'latLon' or 'UTM'
    utm: List
         Names for the Longitude and Latitude columns
         in UTM that are going to be converted.
    lat_lon: List
             Names for the Longitude and Latitude columns
             in Degrees that are going to be converted.
    utm_zone: Int
              UTM Zone code.
    set_south: Bool
               True if located in southern hemisphere,
               False if located in northern hemisphere.

    Returns
    -------

    df: DataFrame
        Inputted GeoDataFrame now with the mileposts
        distance column.

    """

    my_string = "+proj=utm +zone=" + str(utm_zone) + " +ellps=WGS84 +datum=WGS84 +units=m +no_defs"

    if set_south:
        my_string = my_string + ' +south'
    my_proj = Proj(my_string)
    
    if convert_to == 'latLon':
        lon, lat = my_proj(df[utm[0]].values, df[utm[1]].values, inverse=True)
        for c in utm:
            del df[c]
        df[lat_lon[0]] = lat
        df[lat_lon[1]] = lon
    elif convert_to == 'utm':
        x, y = my_proj(df[lat_lon[1]].values, df[lat_lon[0]].values, inverse=False)
        for c in lat_lon:
            del df[c]
        df[utm[0]] = x
        df[utm[1]] = y
    else:
        raise Exception('convertTo must be ["latLon" or "utm"]')
    return df


def flat_geom(geo):

    """
    Flatten GeoDataFrame geometries.
    Ex: All MultiLineStrings become LineStrings.

    Parameters
    ----------
    geo: GeoDataFrame
         GeoDataFrame with the geometries
         that are going to be flatten.

    Returns
    -------
    geo: GeoDataFrame
         GeoDataFrame with flatten geometries

    """

    geo['x'] = geo.apply(fg.getCoords, geom_col="geometry", coord_type="x", axis=1)
    geo['y'] = geo.apply(fg.getCoords, geom_col="geometry", coord_type="y", axis=1)
    
    geo.x = geo.apply(lambda row: [x for x in row['x'] if ~np.isnan(x)], axis=1)
    geo.y = geo.apply(lambda row: [y for y in row['y'] if ~np.isnan(y)], axis=1)
    
    geo['geometry'] = geo.apply(lambda row: LineString([Point(xy) for xy in zip(row.x, row.y)]), axis=1)

    return geo



import locale
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt


def str_format(h, what, per, decimal):
    if per:
        my_str = '{:1.' + str(decimal) + 'f}%'
        return my_str.format(h * 100)
    else:
        label = what.shape[0] * h
        return '{:,.0f}'.format(int(label))


def bars_auto_label(ax, what, per=True, decimal=1, pos_above=False):
    """
    Attach a text label above each bar displaying its height
    """
    # pos is used to divide the height by two
    pos = 1 if pos_above is True else 1  # was 2, acho q está feio no meio
    aux = 0.0005 if pos_above is True else 0.005

    for p in ax.patches:
        if p.get_height() > aux:
            my_text = str_format(p.get_height(), what, per, decimal)
            ax.text(p.get_x() + p.get_width() / 2.,
                    (p.get_y() + p.get_height() / pos),
                    my_text, fontsize=14, ha='center', va='bottom', color='#434853')

    return


def adjust_xlabel(ax, b, flag_lw, flag_up):
    def format_str(val, posi):
        format_num = '%.2f' if (val < 1) and (val > 0.004) else '%d'
        return locale.format(format_num, val, posi)

    xlabels = []
    gap = b[1] - b[0]
    if gap == 1:
        for index, obj in enumerate(b):
            e = format_str(obj, 1)
            xlabels.append(e)

    else:
        for index, obj in enumerate(b):
            e = '[' + format_str(obj, 1) + ', ' \
                + format_str(b[index + 1], 1)
            if index == len(b) - 2:
                # penúltimo item
                e = e + ']'
                xlabels.append(e)
                break
            else:
                e = e + '['
                xlabels.append(e)

    pos = 1 if gap == 1 else 2
    aux = '≤ ' if gap == 1 else '<'
    if flag_lw > 0:
        xlabels[0] = aux + format_str(b[pos - 1], 1)
    if flag_up > 0:
        xlabels[-1] = '≥ ' + format_str(b[-pos], 1)

    # arbitrary delta, bins are always simetric
    # size = (b[2] - b[1]) / 2
    # plt.xticks(b[:-1] + size, np.arange(len(b) - 1))
    ax.set_xticklabels(xlabels, rotation=0, fontsize=12, color='#434853')
    return


def check_what(what):
    if isinstance(what, pd.Series):
        what = pd.DataFrame(what)
    if isinstance(what, pd.DataFrame):
        pass
    else:
        raise Exception('what must be a pd.Series or pd.DataFrame')
    what.columns = [str(w) for w in what.columns]
    what = what.copy(deep=True)
    return what


def make_bins(bins, what):
    # if bins is not a list, make it
    if not (isinstance(bins, list)):
        if isinstance(bins, range):
            bins = list(bins)
        elif isinstance(bins, int):
            # TODO user np.log maybe
            if len(what.columns) > 1:
                i, j = min(what.min()), max(what.max())
                bins = list(np.linspace(start=i, stop=j, num=bins, endpoint=True))
            else:
                s = bins
                val = what.mean()[0]
                bins = []
                for ss in range(s):
                    bins.append(ss * val)
        else:
            raise Exception('bins must be a list, range or integer')
    return bins


def clip2plot(what_df, bins, lwl, upl):
    flag_lw, flag_up = 0, 0
    if lwl < bins[0]: 
        lwl = bins[0]
    if upl > bins[-1]: 
        upl = bins[-1]

    for c in what_df:
        n = what_df.loc[what_df[c].isnull(), c]
        w = what_df.loc[what_df[c].notnull(), c]
        flag_lw = (w < lwl).sum() + flag_lw
        flag_up = (w > upl).sum() + flag_up
        w = np.clip(w, lwl, upl)
        what_df.loc[:, c] = w.append(n)

    return what_df, flag_lw, flag_up


def get_what2plot(what, bins):
    """
    For each column, put value in a given BIN and calc FREQUENCE and (%)
    :param what: dataFrame with series to plot
    :param bins: bins to segregate the data
    :return: dataFrame only with the columns to plot, i.e. PERCENTUAL
    """

    cols2plot = what.columns
    aux = 0
    for c in cols2plot:
        w = what.loc[what[c].notnull(), c]
        mask = what[c].notnull()
        what.loc[mask, 'bin_' + c] = np.digitize(w.astype('float'), bins, right=False)

        # if gap>1:
        # needs to do this...
        what['bin_' + c] = what['bin_' + c] - 1
        what.loc[what[c] == bins[-1], 'bin_' + c] = len(bins) - 1
        aux = -1

        # garantees NaN will be cut out of bins
        mask = what[c].isnull()
        what.loc[mask, 'bin_' + c] = len(bins)
        # calc percentual to plot
        what['frq_' + c] = what.groupby('bin_' + c)['bin_' + c].transform('count')
        what.loc[:, 'per_' + c] = what['frq_' + c] / len(what)

    # organize dataframe to plot
    what2plot = pd.DataFrame(data={'x_pos': list(range(0, len(bins) + aux, 1))})
    for c in cols2plot:
        df = what[['bin_' + c, 'per_' + c]].drop_duplicates()
        df['bin_' + c] = df['bin_' + c].apply(lambda x: x - 1 if x > what2plot.shape[0] - 1 else x)
        what2plot = what2plot.merge(df, left_on='x_pos', right_on='bin_' + c, how='left').fillna(0)
    return what2plot


def redef_cols(what, cols_ref, legenda_txt):
    if (legenda_txt is None) & (len(cols_ref) == 1):
        legenda_bool = False
    else:
        legenda_bool = True

    if legenda_bool:
        if legenda_txt is None:
            legenda_txt = cols_ref
        cols2plot = ['per_' + c for c in cols_ref]
        what = what.rename(columns={cols2plot[0]: legenda_txt})
        cols2plot = legenda_txt
    else:
        cols2plot = ['per_' + c for c in cols_ref]

    return what[cols2plot], legenda_bool


def get_null_stat(what, legenda):
    stat = pd.DataFrame([what.isnull().sum() / len(what), what.notnull().sum() / len(what)])
    if legenda is None: 
        legenda = stat.columns
    stat = stat.rename(columns={stat.columns[0]: legenda})
    stat = stat.T
    stat.columns = ['Ignorados', 'Analisados']
    return stat


def create_figure(title, subtitle, report_nan, source, comentario):
    flatui = ["#C3423F", "#9BC53D", "#fdbf11", "#FDE74C", "#5BC0EB", "#404E4D"]
    sns.set_palette(flatui)
    subtitle = subtitle.capitalize()
    fig = plt.figure(figsize=(14, 7))
    fig.text(0.095, 0.95, title, fontsize=20, **{'fontname': 'Lato'})
    fig.text(0.095, 0.9, subtitle, fontsize=14, **{'fontname': 'Lato'})
    if source != '':
        fig.text(0.095, 0.030, "Fonte: ", weight='bold', fontsize=12, **{'fontname': 'Lato'})
        fig.text(0.145, 0.030, f"{source}", fontsize=12, **{'fontname': 'Lato'})
    else:
        pass
    if comentario != '':
        fig.text(0.095, 0, f"{comentario}", fontsize=12, **{'fontname': 'Lato'})
    else:
        pass

    # config grid and axis
    sns.set_style("whitegrid")

    if report_nan:
        grid_size = (1, 10)
        ax1 = plt.subplot2grid(grid_size, (0, 1), colspan=9)  # plot what
        ax2 = plt.subplot2grid(grid_size, (0, 0), colspan=1)  # plot stat

        ax1.xaxis.grid(False)
        ax2.xaxis.grid(False)
    else:
        ax1 = fig.subplots()  # plot what
        ax1.xaxis.grid(False)
        ax2 = None

    # ax1.axis('off')
    # ax2.axis('off')
    sns.despine(left='False')
    # for item in [fig, ax1, ax2]:
    #    item.patch.set_visible(False)

    return fig, ax1, ax2


def hide_grid(ax):
    ax.yaxis.grid(False)  # horizontal lines
    ax.xaxis.grid(False)  # vertical lines
    ax.set_yticks([])  # to erase y labels
    return



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 12 10:11:32 2022

@author: bcalazans
"""
import os
import webbrowser


def text2html(title=False, subtitle=False, text=[]):
    '''
    Transforma strings de texto para formato em HTML

    Parameters
    ----------
    title : STRING, optional
        Texto a ser formatado como título. The default is False.
    subtitle : STRING, optional
        Texto a ser formatado como subtítulo. The default is False.
    text : STRING ou lista de STRINGS, optional
        Textos padrão. Força uma nova linha a cada item da lista. 
        The default is [].

    Returns
    -------
    HTML string
        String de texto contendo os parâmetros passados formatados para HTML.

    '''
    if not isinstance(text, list): text = [text]
    
    html = '<div>'
    if title: html = html + '<h1>'+title + '</h1>'
    if subtitle: html = html + '<h2>'+subtitle + '</h2>'
    # each member list is a new line
    #text = [t +'<br>' for t in text]
    html = html + '<p style="color:#1F1321;font-size:20px;">'+ '<br>'.join(text) + '</p>'  
 
    return html + '</div>'


def gera_doc_html(list2plot):
    '''
    Dado uma lista de items, gera uma super string formatada em HTML
    
    list2plot : LIST or DICT
        Lista de elementos para salvar no HTML.
        Esses elementos podem ser:
            * uma string já formatada HTML (use a função text2html para isso)
            * uma figura do plotly em HTML
            * uma figura do matplotlib : rasteriza automaticamente para html
            * uma lista com duas posições de figuras HTML que serão salvas lado a lado
    '''
    html_str_complete = '<!DOCTYPE html><html><div>'
    for fig in list2plot:
        if isinstance(fig, str):
            # já é um texto
            html_string = fig
        elif isinstance(fig, list): 
            # list of two figures to put side by side in a simple way
            str1= fig[0].to_html(full_html=False, include_plotlyjs='cdn')
            str1 = str1.replace('<div>', '<div style="width: 45%; float:left">')
             
            str2= fig[1].to_html(full_html=False, include_plotlyjs='cdn')
            str2 = str2.replace('<div>', '<div style="width: 45%; float:right">')
             
            html_string = '<div>' + str1 + str2 + '</div>'
        else:
            # é uma figura do plotly, passar para texto html
            html_string = fig.to_html(full_html=False, include_plotlyjs='cdn')
            #html_string = html_string.replace('width:100%','width: 80%, align="center"')
        html_str_complete = html_str_complete + html_string 
    html_str_complete + '</div></html>'      
    
    return html_str_complete


def save_html(path, name, list2plot):
    '''
    Salva um HTML com os elementos de list2plot
        
    Parameters
    ----------
    path : STRING
        Caminho onde salvar o arquivo
    name : STRING
        Nome com extensão .html para salvar o arquivo
     
    list2plot: LIST ou DICT
        Caso receba um DICT, cria um HTML com abas em que cada KEY é uma aba.
        {nome_aba:[item1, item2, [item3, item4]]}
        Lista de elementos para salvar no HTML.
        Esses elementos podem ser:
            * uma string já formatada HTML (use a função text2html para isso)
            * uma figura do plotly em HTML
            * uma figura do matplotlib : rasteriza automaticamente para html
            * uma lista com duas posições de figuras HTML que serão salvas lado a lado

    Returns
    -------
    None

    '''
    
    
    html_str= ''
    if isinstance(list2plot, list): 
        html_str = gera_doc_html(list2plot)
    elif isinstance(list2plot, dict):
        html_str+=html_header
        i=0
        for key, items in list2plot.items():
            i+=1
            # html string que define as abas
            html_aba = f'''
            <li>
            <input type="radio" name="tabs" class="rd_tabs" id="tab{i}" checked>
            <label for="tab{i}">{key}</label>
            <div class="content">
            <article>
            '''
            html_str+=html_aba
            # gera string com todos os gráficos            
            html_str+=gera_doc_html(items)
            # fecha divisória das abas
            html_str+='</article></li>'
            #html_str+='</article></div></li>'
        html_str+='</ul></nav></body></html>'
    else:
        raise Exception('list2plot must be a LIST or a DICT where the items are lists')
        
    # open file and write the HTML string        
    filename = os.path.join(path, name)
    with open(filename, 'w') as file:
        file.write(html_str)

    return         
  
    
def open_html(path, name):
    filename = "file:///" + os.path.join(path, name)
    try: webbrowser.open_new_tab(filename)
    except: print (f'ERROR trying to open {name:}')
    return  


html_header = '''
<!DOCTYPE html>
<html>
<head>
<title>Modelos de Regressão</title>
<style>
*{
padding: 0;
}
body{
background-color: #fff;
font-family: Arial;
}
.nav_tabs{
margin: 5px auto;
background-color: #fff;
position: relative;
}
.nav_tabs ul{
list-style: none;
}
.nav_tabs ul li{
float: left;
height: 320px;
width: 165px;
}
.nav_tabs label{
width: 120px;
height: 25px;
padding: 20px;
border-radius: 8px 8px 0 0;
background-color: #363b48;
display: block;
color: #fff;
cursor: pointer;
text-align: center;
}
.rd_tabs:checked ~ label{
background-color: #e54e43;
}
.rd_tabs{
display: none;
}
.content{
border-top: 5px solid #e54e43;
background-color: #fff;
display: none;
position: absolute;
width: 1000px;
left: 0;
}
.rd_tabs:checked ~ .content{
display: block;
}

</style>
</head>
<body>

  <nav class="nav_tabs">
      <ul>
         '''



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 30 08:32:27 2020

@author: bcalazans

Funções para plotar resultados da regressão em um arquivo HTML
"""
import os, sys
import pandas as pd
import scipy as sp
import numpy as np


import matplotlib.pyplot as plt
import statsmodels.stats.api as sms
import statsmodels.stats.stattools as st
from statsmodels.stats.outliers_influence import variance_inflation_factor
from statsmodels.tsa.stattools import acf

from plotly.subplots import make_subplots
import plotly.graph_objects as go
import plotly.express as px
import plotly.figure_factory as ff
import plotly as py


# systool modules
# TODO ideally i do not want to mess around with path
sys.path.insert(1, os.path.abspath('.'))

def text2html(title=False, subtitle=False, text=[]):
    '''
    Transforma strings de texto para formato em HTML

    Parameters
    ----------
    title : STRING, optional
        Texto a ser formatado como título. The default is False.
    subtitle : STRING, optional
        Texto a ser formatado como subtítulo. The default is False.
    text : STRING ou lista de STRINGS, optional
        Textos padrão. Força uma nova linha a cada item da lista.
        The default is [].

    Returns
    -------
    HTML string
        String de texto contendo os parâmetros passados formatados para HTML.

    '''
    if not isinstance(text, list): text = [text]

    html = '<div>'
    if title: html = html + '<h1>' + title + '</h1>'
    if subtitle: html = html + '<h2>' + subtitle + '</h2>'
    # each member list is a new line
    # text = [t +'<br>' for t in text]
    html = html + '<p style="color:#1F1321;font-size:20px;">' + '<br>'.join(text) + '</p>'

    return html + '</div>'

#%% nomeia arquivo HTML no formato desejado 
#regModel_Ycolumns=Xcolumns_intercepYESorNO_ZNnumberOfRows.html
def concatena_cols(df):
    columns = list(df.columns) if len(df.shape) > 1 else df.name
    tam = len(columns)
    palavra = ''
    for pos in range(tam):
        if(pos<(tam-1)):
            palavra = palavra + columns[pos] + '+'
        else:
            palavra = palavra + columns[pos]
    return palavra

# Monta o nome do HTML
def give_name(x, y, intercept):
    Ycolumns = y.name
    Xcolumns = concatena_cols(x)
    if(intercept):
        inter = '_intercepYES_ZN'
    else:
        inter = '_intercepNO_ZN'
    number_of_rows = str(len(x))
    return 'regModel_' + Ycolumns + '=' + Xcolumns + inter + number_of_rows + '.html'


#%% Aux plots
def plot_predictions(y, y_predicted, student_residuals):
    '''Plota lado a lado gráfico de Observed X Predicted + Studentized Errors'''
    
    fig = make_subplots(rows=1, cols=2, subplot_titles=['Valores observados X estimados',
                                                   'Scale Location'])
    
    # Observed X Predicted
    X_plot = [min(y), max(y)]
    fig.add_trace(go.Scatter(x=y, y=y_predicted, 
                             #name="pontos", 
                             mode='markers', 
                             marker=dict(color='darkblue', size=8),
                             hovertext=y.index),
                 row=1, col=1)
    fig.add_trace(go.Scatter(x=X_plot, y=X_plot, line=dict(color='darkred', width=3)),
                 row=1, col=1)
    fig.update_xaxes(title_text="Observado", row=1, col=1)
    fig.update_yaxes(title_text="Estimado", row=1, col=1)
    
    # Scale-location
    fig.add_trace(go.Scatter(x=y_predicted, y=student_residuals, 
                             mode='markers', 
                             marker=dict(color='darkblue', size=8)),
                 row=1, col=2)
    fig.update_xaxes(title_text="Estimado", row=1, col=2)
    fig.update_yaxes(title_text="Studentized Residuals", row=1, col=2)
    fig.update_layout(showlegend=False)
    return fig


def residuals_histogram(residuals):
    #this works
    '''
    fig = go.Figure(
        layout=dict(
            title="Residuals Histogram",
            xaxis_title='Residuals',
            yaxis_title='',
        )
    )
    
    #fig = go.Figure(data=[go.Histogram(x=residuals, histnorm='probability')])
    
    fig.add_trace(
            go.Histogram(x=residuals,  histnorm='density', xbins=dict(size=10))
    )
    '''
    # this also works
    #fig = px.histogram(x=residuals, nbins=10, title='Histogram of the Residuals')
    fig, (aq, ax) = plt.subplots(1,2, figsize=(16,5))
    _, (__, ___, r) = sp.stats.probplot(residuals, plot=aq, fit=True)
    aq.spines['top'].set_visible(False)
    aq.spines['right'].set_visible(False)

    # Create distplot in matplolib (there is no kde func in plotly)
    #fig, ax = plt.subplots()
    ax=residuals.plot.kde(ax=ax, title='Histograma dos residuais')
    mean_val = np.mean(ax.get_children()[0]._x)
    std_val = np.std(ax.get_children()[0]._x)
    max_val = np.max(ax.get_children()[0]._y)
    # Annotate points
    ax.annotate('{:10}{:>10,.2f}<br>{:10}{:>10,.2f}'.format('  Média',mean_val,'  Desvio',std_val), 
                xy=(mean_val, max_val),verticalalignment='top')
    # vertical dotted line originating at mean value
    #plt.axvline(mean_val, linestyle='dashed', linewidth=2, color='#202231') # does not work with plotly
    ax.plot([mean_val,mean_val], [0,max_val], linestyle='dashed', linewidth=2, color='#202231')
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    
    # plot histogram
    ax=residuals.hist(density=1, ax=ax, grid=False)
        
    #transform in plotly object figure
    fig = py.tools.mpl_to_plotly(fig)
    
    return fig


def plot_harvey(PVALUE):
    fig = go.Figure(go.Indicator(
    domain = {'x': [0, 1], 'y': [0, 1]},
    value = PVALUE,
    number = {'suffix': "%"},
    mode = "gauge+number",
    title = {'text': "Harvey-Collier", 'font': {'size': 24}},
    #delta = {'reference': 380},
    gauge = {'axis': {'range': [0, 100]},
             'bar': {'thickness': 0},
             'steps' : [
                 {'range': [0, 5], 'color': "darkred"},
                 {'range': [5, 100], 'color': "green"}],
             'threshold' : {
                 'line': {'color': "black", 'width': 4}, 'value': PVALUE}}))
    return fig 


def tablePlot(df_table, index=False):
    '''
    fig = go.Figure(
        layout=dict(
            title="Homoscedasticity"
        )
    )
    fig.add_trace(
            go.Table(
                    header=dict(
                            values = names,
                            font = dict(size=10),
                            align = "left"),
                    cells=dict(
                            values = param,
                            align = "left")
            )
    )
    '''
    
    fig =  ff.create_table(df_table, index=index)
    fig.update_layout(title="Durbin-Watson",hovermode=False, width=800)
    return fig


def durbin_watson(pvalue):
    '''
    fig = go.Figure(go.Indicator(
    domain = {'x': [0, 1], 'y': [0, 1]},
    value = PVALUE,
    mode = "gauge+number+delta",
    title = {'text': "Durbin-Watson", 'font': {'size': 24}},
    #delta = {'reference': 380},
    gauge = {'axis': {'range': [0, 4]},
             'bar': {'thickness': 0},
             'steps' : [
                 {'range': [0, 2], 'color': "darkred"},
                 {'range': [2, 4], 'color': "darkblue"}],
             'threshold' : {
                 'line': {'color': "black", 'width': 4}, 'value': PVALUE}}))
    '''
    #pvalue=1.437
    pvalue = round(pvalue, 4)
    img = np.tile(np.linspace(0,4, num=200),(2,1))
    #img = [np.linspace(0,4, num=100)]

    fig = make_subplots(specs=[[{"secondary_y": True}]])

    #fig.add_trace(px.imshow(img,color_continuous_scale ='RdBu').data[0], row=1, col=1)
    #fig = px.imshow(img,color_continuous_scale ='RdBu')
    fig.add_trace(go.Heatmap(z=img, colorscale='RdBu',showscale =False), 
              secondary_y=False)

    fig.add_trace(go.Scatter(
                    x=[pvalue*50, pvalue*50],
                    y=[0, 2],
                    mode="lines+text",
                    line=go.scatter.Line(color="black"), 
                    showlegend=False,hoverinfo='skip'),
    secondary_y=True
    )
    fig.update_layout(
    annotations=[
        dict(
            x=pvalue*50-4,
            y=0.55, textangle=-90,
            showarrow=False,
            text=str(pvalue))])
    fig.update_yaxes(title_text="Correlação <b>negativa</b>", showticklabels=False, secondary_y=False)
    fig.update_yaxes(title_text="Correlação <b>positiva</b>", showticklabels=False, secondary_y=True)

    fig.update_xaxes(title_text="",
                 tickmode="array",tickvals =[0,50,100,150,199],ticktext=['0','1','2','3','4'])

    fig.update_layout(title="Durbin-Watson",hovermode=False, height=330)
    fig.show()
    return fig
    

def prepare_data_corr_plot_alterado(x, lags, zero):
    zero = bool(zero)
    irregular = False if zero else True
    if lags is None:
        # GH 4663 - use a sensible default value
        nobs = x.shape[0]
        lim = min(int(np.ceil(10 * np.log10(nobs))), nobs - 1)
        lags = np.arange(not zero, lim + 1)
    elif np.isscalar(lags):
        lags = np.arange(not zero, int(lags) + 1)  # +1 for zero lag
    else:
        irregular = True
        lags = np.asanyarray(lags).astype(np.int)
    nlags = lags.max(0)

    return lags, nlags, irregular
    
    

def plot_acf_durbin_watson(pvalue, x, ax=None, lags=None, alpha=.05, 
                           use_vlines=True, qstat=False,fft=False, 
                           zero=True,
                           vlines_kwargs=None, **kwargs):
    
    fig = make_subplots(rows=1, cols=2, subplot_titles=['Autocorrelation','Durbin Watson = ' + '{:.5}'.format(pvalue)], 
                        specs=[[{},{"secondary_y": True}]])

    # PLOT ACF
    #fig, ax = utils.create_mpl_ax(ax)
    
    lags, nlags, irregular = prepare_data_corr_plot_alterado(x, lags, zero)
    vlines_kwargs = {} if vlines_kwargs is None else vlines_kwargs
    
    nlags=len(x)-1
    lags = np.arange(not zero, nlags + 1)
    
    confint = None
    # acf has different return type based on alpha
    if alpha is None:
        acf_x = acf(x, nlags=nlags, alpha=alpha, fft=fft,qstat=qstat)
    else:
        acf_x, confint = acf(x, nlags=nlags, alpha=alpha, fft=fft,qstat=qstat)
    
    if irregular:
        acf_x = acf_x[lags]
        if confint is not None:
            confint = confint[lags]

    lags_original = lags
    acf_x_original = acf_x

    if confint is not None:
        if lags[0] == 0:
            lags = lags[1:]
            confint = confint[1:]
            acf_x = acf_x[1:]
        lags = lags.astype(np.float)
        lags[0] -= 0.5
        lags[-1] += 0.5
        #ax.fill_between(lags, confint[:, 0] - acf_x,confint[:, 1] - acf_x, alpha=.25)
        
        fig.add_trace(go.Scatter(x=lags, y=confint[:, 0] - acf_x,
                        fill=None,
                        mode='lines',
                        line_color='rgb(234, 188, 196)'), row=1, col=1)
        fig.add_trace(go.Scatter(x=lags,
                        y=confint[:, 1] - acf_x,
                        fill='tonexty', # fill area between trace0 and trace1
                        mode='lines', line_color='rgb(234, 188, 196)'), row=1, col=1)       
        fig.update_layout(xaxis_showgrid=False, yaxis_showgrid=False, xaxis_zeroline=False, showlegend=False)#plot_bgcolor='white'
        
        fig.add_trace(go.Scatter(x=lags_original, y=acf_x_original, mode='markers', marker=dict(color='darkblue', size=6)), row=1, col=1)
        #fig.show()
        
    # PLOT DURBIN WATSON
    
    pvalue = round(pvalue, 4)
    img = np.tile(np.linspace(0,4, num=200),(2,1))
    #img = [np.linspace(0,4, num=100)]
    
    #fig = make_subplots(specs=[[],[{"secondary_y": True}]])
    
    #fig.add_trace(px.imshow(img,color_continuous_scale ='RdBu').data[0], row=1, col=1)
    #fig = px.imshow(img,color_continuous_scale ='RdBu')
    #fig.add_trace(go.Heatmap(z=img, colorscale='RdBu',showscale =False), secondary_y=False, row=1, col=2)
    fig.add_trace(go.Heatmap(z=img, colorscale='RdBu',showscale =False), row=1, col=2)
    fig.add_trace(go.Scatter(
                    x=[pvalue*50, pvalue*50],
                    y=[0, 2],
                    mode="lines+text",
                    line=go.scatter.Line(color="black"), 
                    showlegend=False,hoverinfo='skip'),
    secondary_y=True, row=1, col=2)
    fig.update_layout(
        annotations=[
            dict(
                x=pvalue*50-4,
                y=0.55, textangle=-90,
                showarrow=False,
                text=str(pvalue))])
    
    fig.update_yaxes(title_text="Correlação <b>negativa</b>", showticklabels=False, secondary_y=False,row=1, col=2)
    fig.update_yaxes(title_text="Correlação <b>positiva</b>", showticklabels=False, secondary_y=True,row=1, col=2)
    '''
    
    fig.update_yaxes(title_text="Correlação <b>negativa</b>", showticklabels=False, row=1, col=2)
    fig.update_yaxes(title_text="Correlação <b>positiva</b>", showticklabels=False, row=1, col=2)
    '''
    fig.update_xaxes(title_text="",tickmode="array",tickvals =[0,50,100,150,199],ticktext=['0','1','2','3','4'], row=1, col=2)

    #fig.update_layout(title="Durbin-Watson",hovermode=False, height=330, row=1, col=2)
    
    return fig

                    

def multicollinearity(corr_df):
        '''
        fig = go.Figure(data=go.Heatmap(z=val_x.corr().values.tolist()),
                    x = list(val_x.columns),
                    y = list(val_x.columns))
        #fig.add_trace(go.Heatmap(x))
        fig.update_layout(title='Multicollinearity of features')
        '''
        # good color scales: Blues, bluyl, brwnyl, bugn, bupu, ylorrd
        fig = px.imshow(corr_df, color_continuous_scale ='Blues', 
                        title='Multicollinearity of features'
                        )
        
        return fig


def write_formula_fancy(model, y):
    formula = ''
    for var, coef in zip(reversed(model.params.index.tolist()), reversed(model.params.values)):
        if var =='const': formula = formula + '{0:+.3f}'.format(coef)
        else: formula = formula + '{0:+.3f}'.format(coef) + ' * ' + var + ' '
    formula = y.name + ' = ' + formula
    return formula

#%% main plotter
def plot_regression(model, x,y, intercept):
    list_2_plots = []
    residuals = model.resid
        
    # Escreve equação no início do documento        
    formula = write_formula_fancy(model, y)
    list_2_plots.append(
        text2html(title='REGRESSÃO LINEAR - ' + y.name + ' - ' + str(len(y)) + ' ZONAS',
                  subtitle=formula,
                  text='R-squared {:10.4}'.format(model.rsquared)))
    
    # see exog vars
    #for X_var in x.columns: _ = sm.graphics.plot_regress_exog(model, X_var)

    ## Observed vs Predicted values and Studentized residuals         
    student_residuals = pd.Series(np.abs(model.get_influence().resid_studentized_internal))
    list_2_plots.append(
        plot_predictions(y, model.fittedvalues, student_residuals))
    
    list_2_plots.append(
        text2html(text=[
                   'Para cada variável dependente, observar se:',
                   '   * p-value < 5%',
                   '   * t-value > 2',
                   'Caso contrário, tente retirar a variável da regressão para melhorar o modelo']
            ))
    
    values = pd.DataFrame([model.pvalues,model.tvalues], index=['p-val', 't-val']).T
    
    values['status'] = np.where((values['p-val'] < 0.05) & (values['t-val'] > 2),
                                'Passed!','A regressão deve melhorar sem essa variável... =/')
    values['p-val'] = values['p-val'].map('{:,.4%}'.format)
    values['t-val'] = values['t-val'].map('{:,.4f}'.format)
    #values['p-val'].style.applymap(lambda val: 'color:red' if val < 0 else 'black'])
    list_2_plots.append(tablePlot(values, index=True))
    
    ## TEST ASSUMPTIONS TO REGRESSION
    # 1)LINEARITY: Harvey-collier multiplier and Rainbow  
    list_2_plots.append(
        text2html('TEST ASSUMPTIONS TO REGRESSION',
                  '1# LINEARITY',
                  ['''This test performs a t-test with parameter degrees of freedom on the recursive residuals. 
                   Recursive residuals are linear transformations of ordinary residuals and are independently
                   and identically distributed. If the true relationship is not linear but convex or concave the mean of 
                   the recursive residuals should differ from zero significantly. A statistically significant result (pvalue>0,05) means
                   that we can reject the null hypothesis of the true model being linear.''']
                  ))
    
    # harvey dá erro se enviar SingularMatrix
    try: list_2_plots.append(plot_harvey(sms.linear_harvey_collier(model).pvalue * 100))
    except: list_2_plots.append(plot_harvey(0))
    #sm.stats.linear_rainbow(model)   # indicate whether the linear fit of the model is adequate even if some underlying relationships are not linear
    
    # Harvey, we will accept linearity if p-value > 5%
    # The Harvey-Collier test performs a t-test (with parameter degrees of freedom) on the recursive residuals. If the true relationship is not linear but convex or concave the mean of the recursive residuals should differ from 0 significantly.
    # HARVEY: indicates whether the residuals are linear
    '''
    print('HARVEY IS HAPPY =D') if sms.linear_harvey_collier(model).pvalue > 0.05 else print('HAVEY IS SAD =/')
    linearity_test = pd.DataFrame([sms.linear_harvey_collier(model).pvalue], 
                     index=['Harvey-collier (residuals teste)'])
    linearity_test.columns = ['p-values']
    linearity_test['p-value %'] = linearity_test.applymap(lambda x: '{0:.4%}'.format(x))
    linearity_test['t-value'] = model.tvalues
    list_2_plots.append(tablePlot(linearity_test, index=True))
    '''
    list_2_plots.append(
        text2html('',
                  '2# NORMALITY',
                  ['Os pontos longe da reta no QQ-plot indicam outliers ou que a regressão não consegue explicá-los muito bem',
                   'No histograma, esperamos ver a bell-curve dos residuais com média 0 e variância 1'])
        )
    
    # 2) NORMALITY: check if residuals are normally distributed
    # Plot residual Q-Q plot
    # Plot Histogram of residuals distribution
    # If the residuals are normally distributed, we should see a bell-shaped histogram centred on 0 and with a variance of 1.
    list_2_plots.append(residuals_histogram(residuals))

    #TODO - STEFANY URGENTE ERRO NA LINHA ACIMA

    # TODO chekcing OMnibus...Bruna entender isso
    #Omnibus is a test of the skewness and kurtosis of the residual. A high value indicates a skewed distribution whereas a low value (close to zero) would indicate a normal distribution
    #The Prob (Omnibus) performs a statistical test indicating the probability that the residuals are normally distributed. We hope to see something close to 1.
    
    list_2_plots.append(
        text2html('',
                  '3# HOMOSCEDASTICITY',
                  ['residuals should be equal for all predicted dependent variables scores',
                   'Breush-Pagan test: measures how errors increase across the explanatory variable',
                   'hight values of Legrange indicates presence of heteroskedasticity (i.e. bad!)'])
        )
    # 3) HOMOSCEDASTICITY: residuals should be equal for all predicted dependent variables scores
    # Breush-Pagan test: measures how errors increase across the explanatory variable        
    names=['Lagrange multiplier statistic', 'p-value','f-value', 'f p-value']
    param=list(sms.het_breuschpagan(model.resid, model.model.exog))      
    # hight values of Legrange indicates presence of heteroskedasticity (i.e. bad!)
    list_2_plots.append(tablePlot(pd.DataFrame(data={'Indicadores':names, 
                                                      'Valores': param})))
    
    # check here if p-value < 5% then Homoscedasticity is FALSE
    #TODO se descobrir range do Lagrange multiplier, colcoar um hodometro Bruna
    
    list_2_plots.append(
        text2html('',
                  '4# INDEPENDENCE',
                  ['Check independence of residuals. The less correleted, the better.'])
        )
    
    # 4) INDEPENDENCE
    # ACF plot - every row should be under the blue box  
    # TODO - ACF plot
    # TODO colocar ACF plot e DurbinWatson lado a lado
    #_ = smt.graphics.plot_acf(model.resid, alpha=0.05)
    
    # TENTANDO JUNTAR ACF COM DURBIN-WATSON


    # Durbin-Watson : check independence of the residuals
        # value = 2 means that there is no autocorrelation in the sample,
        # values < 2 indicate positive autocorrelation,
        # values > 2 negative autocorrelation.
        #The Durbin-Watson statistic will always have a value between 0 and 4. A value of 2.0 means that there is no autocorrelation detected in the sample. Values from 0 to less than 2 indicate positive autocorrelation and values from from 2 to 4 indicate negative autocorrelation.
    print('Durbin-Watson',st.durbin_watson(residuals, axis=0))
    PVALUE = st.durbin_watson(residuals, axis=0)
    #list_2_plots.append(durbin_watson(PVALUE))
    list_2_plots.append(plot_acf_durbin_watson(PVALUE, model.resid, alpha=0.05))
    # If has more then one column, check MULTICOLLINEARITY
    if len(x.columns) > 1:        
        list_2_plots.append(
            text2html('',
                      '5# MULTICOLLINEARITY',
                      ['Variáveis devem ter pouca correlação entre si',
                       'VIF of 5 or 10 and above indicates a multicollinearity problem'])
            )

        # 5) MULTICOLLINEARITY
        # VIF of 5 or 10 and above indicates a multicollinearity problem
        # For each X, calculate VIF and save in dataframe
        
        #A VIF of over 10 for some feature indicates that over 90\% of the variance in that feature is explained by the remaining features. Over 100 indicates over 99\%.
        
        df_vif = pd.DataFrame(index=x.columns)
        df_vif["VIF Factor"] = [variance_inflation_factor(x.values, i) for i in range(x.shape[1])]
        df_vif["VIF Factor"] = df_vif["VIF Factor"].map('{:,.4f}'.format)
        list_2_plots.append(tablePlot(df_vif, index=True))
        
        #another ways is to see teh correlation directly                
        list_2_plots.append(multicollinearity(x.corr()))
        

        # TODO colocar a tabela do VIF ao lado da grafico de multicorrlinear
        # TODO salvar o sns.pairplot como PNG e colcoar ele no HTML
        #sns.pairplot(data =x)  
 
    return list_2_plots



# -*- coding: utf-8 -*-
"""
Created on 08-08-2019

@author: bcalazans

Script com as funções necesarias para achatar geometrias
USado principalmente para garantir que multilinestrings passadas sejam
convertidas em lineString

MainFunction: flat_geom
"""
import numpy as np
import shapely

def getXYCoords(geometry, coord_type):
    """ Returns either x or y coordinates from  geometry coordinate sequence. Used with LineString and Polygon geometries."""
    if coord_type == 'x':
        return geometry.coords.xy[0]
    elif coord_type == 'y':
        return geometry.coords.xy[1]


def getPolyCoords(geometry, coord_type):
    """ Returns Coordinates of Polygon using the Exterior of the Polygon."""
    ext = geometry.exterior
    return getXYCoords(ext, coord_type)


def getLineCoords(geometry, coord_type):
    """ Returns Coordinates of Linestring object."""
    return getXYCoords(geometry, coord_type)


def getPointCoords(geometry, coord_type):
    """ Returns Coordinates of Point object."""
    if coord_type == 'x':
        return geometry.x
    elif coord_type == 'y':
        return geometry.y

    
def multiGeomHandler(multi_geometry, coord_type, geom_type):
    """ 
    Function for handling multi-geometries. Can be MultiPoint, MultiLineString or MultiPolygon. 
    Returns a list of coordinates where all parts of Multi-geometries are merged into a single list. 
    Individual geometries are separated with np.nan which is how Bokeh wants them. 
    # Bokeh documentation regarding the Multi-geometry issues can be found here (it is an open issue)
    # https://github.com/bokeh/bokeh/issues/2321
    """
    
    for i, part in enumerate(multi_geometry):
        # On the first part of the Multi-geometry initialize the coord_array (np.array)
        if i == 0:
            if geom_type == "MultiPoint":
                coord_arrays = np.append(getPointCoords(part, coord_type), np.nan)
            elif geom_type == "MultiLineString":
                coord_arrays = np.append(getLineCoords(part, coord_type), np.nan)
            elif geom_type == "MultiPolygon":
                coord_arrays = np.append(getPolyCoords(part, coord_type), np.nan)
        else:
            if geom_type == "MultiPoint":
                coord_arrays = np.concatenate([coord_arrays, np.append(getPointCoords(part, coord_type), np.nan)])
            elif geom_type == "MultiLineString":
                coord_arrays = np.concatenate([coord_arrays, np.append(getLineCoords(part, coord_type), np.nan)])
            elif geom_type == "MultiPolygon":
                coord_arrays = np.concatenate([coord_arrays, np.append(getPolyCoords(part, coord_type), np.nan)])
    
    # Return the coordinates 
    return coord_arrays
    

def getCoords(row, geom_col, coord_type):
    """
    Returns coordinates ('x' or 'y') of a geometry (Point, LineString or Polygon) as a list (if geometry is LineString or Polygon). 
    Can handle also MultiGeometries.
    """
    # Get geometry
    geom = row[geom_col]
    
    # Check the geometry type
    gtype = geom.geom_type
    
    # "Normal" geometries
    # -------------------
    
    if gtype == "Point":
        return getPointCoords(geom, coord_type)
    elif gtype == "LineString":
        return list( getLineCoords(geom, coord_type) )
    elif gtype == "Polygon":
        return list( getPolyCoords(geom, coord_type) )
        
    # Multi geometries
    # ----------------
    
    else:
        return list( multiGeomHandler(geom, coord_type, gtype) ) 
    

def flat_geom(geo):
    '''
    Recebe uma geoPandasDataFrame com geometrias e flatten them
    Ex. todas as multilinestrings viram LineStrings
    '''
    geo['x'] = geo.apply(getCoords, geom_col="geometry", coord_type="x", axis=1)
    geo['y'] = geo.apply(getCoords, geom_col="geometry", coord_type="y", axis=1)
    
    geo.x=geo.apply(lambda row: [x for x in row['x'] if ~np.isnan(x)],axis=1)
    geo.y=geo.apply(lambda row: [y for y in row['y'] if ~np.isnan(y)],axis=1)
    
    geo['geometry'] = geo.apply(lambda row: shapely.geometry.LineString([shapely.geometry.Point(xy) for xy in zip(row.x,row.y)]),axis=1)
    return geo
    


# -*- coding: utf-8 -*-
"""
Created on Tue Apr 12 10:28:14 2022

@author: bcalazans & MoleDownTheHole (pcardoso)

Módulo com funções úteis no tratamento de dados tabulados.

"""
# Python Modules
import pandas as pd

# Systools Modules
from .utils import readWrite as rW


def dataframe2numeric(df, col_dt_preffix='', col_td_preffix=''):

    """
    Format the DataFrame columns to Numeric and Time Formats.
    Support datetime, timestamp, int and float formats.

    Parameters
    ----------
    df : DataFrame
        DataFrame type variable that has columns who will be converted
        to numeric or time format.

    col_dt_preffix : string, optional
        String type variable that have the prefixes used to show that
        the designated column title that contains it,
        must be converted to datetime type.

    col_td_preffix : string, optional
        String type variable that have the prefixes used to show that
        the designated column title that contains it,
        must be converted to timedelta type.

    Returns
    -------

    df : DataFrame
    DataFrame used in the input, but with the numeric and time columns converted.

    """

    other_cols = [col for col in df.columns if df[col].dtype != 'datetime64[ns]']
    all_cols = df.columns
    
    # Transform to numeric but keep datetime.
    df[other_cols] = df[other_cols].apply(pd.to_numeric, errors='ignore')
    
    # Select Cols to Convert
    if col_dt_preffix != '':

        cols_dt = [col for col in all_cols if str(col).startswith(col_dt_preffix)]

    else:
        cols_dt = []

    if col_td_preffix != '':

        cols_td = [col for col in all_cols if str(col).startswith(col_td_preffix)]

    else:
        cols_td = []
    # Transform Timedeltas and Datetimes to its format.
    for col in cols_td:
        df[col] = pd.to_timedelta(df[col])

    # Transform in datetime commonly used formats in Brazil.

    time_formats = ["%d/%m/%Y %H:%M:%S",
                    "%d/%m/%Y %H:%M",
                    "%Y/%m/%d %H:%M:%S",
                    "%Y-%m-%d %H:%M:%S",
                    "%Y-%m-%d %H:%M"]

    for col in cols_dt:
        for time_format in time_formats:
            try:
                df[col] = pd.to_datetime(df[col], format=time_format)
                break
            except ValueError:
                continue

        if df[col].dtype != 'datetime64[ns]':
            df[col] = pd.to_datetime(df[col], infer_datetime_format=True)

    return df


def open_file(path, name=None, expected_cols=None, usa=False, col_td_prefix='gap',  col_dt_prefix='H_', **kwargs):

    """
    Open file as a DataFrame.
    Support csv, txt, parquet, excel and shape.

    Parameters
    ----------
    path : string
        Path of the file, may include the file name with extension.
        If the file is zipped, pass the path with the zip extension.

    name : string, optional
        Name of the file with extension to be open.
        Expects that the name was passed on the path arg.

    expected_cols : list, optional
        Import only this cols.
    usa : Bool, optional
        If TRUE expects comma separator and point as decimal indicator.
        If FALSE expects ; as separator and comma as decimal indicator.

    col_dt_prefix: String, optional
        Indicates column name prefix to be converted to datetime.

    col_td_prefix: String, optional
        Indicates column name prefix to be converted to timedelta.

    **kwargs : dict
        Any argument for the original read functions from (geo)pandas.

    Returns
    -------
    df : pd.DataFrame or geo.DataFrame
        DataFrame do arquivo lido contendo ou não geometria.

    """
    
    df = rW.load_file(path, name, expected_cols, usa, **kwargs)
         
    # Configure Formats
    df = dataframe2numeric(df, col_dt_prefix, col_td_prefix)
    return df


# Insert excel parameters if needed, after the "truncate_sheet" parameter.

def save_file(df, path, name='test', usa=False, ext='csv',
              sheet_name='Sheet1', start_row=None, truncate_sheet=False):
    """
    Save DataFrame to an exterior file created and formated based
    on various parameters.
    Support csv, txt, parquet, excel and shape.

    Parameters
    ----------
    df : DataFrame
        DataFrame type variable that has columns who the user
        wants to export.

    path : String
        Path of the file.
        Doesn't need to have the filename and extension included.

    name : String, optional
        File name.
        Doesn't need to have the extension included.

    usa : Boolean, optional
        True -> The exported file uses the american format.
        (Ex: Comma as separator, dot as decimal...)
        False -> The exported file doesn't uses the american format.
        (Ex: Semicolon as separator, comma as decimal...)

    ext : String, optional
        The file extension.
        Must be csv, txt, parquet, excel or shape.

    sheet_name : String, optional
        The file, when saved as an excel file, sheet name.

    start_row : Integer, optional
        The excel file row where the user wants the new data to be inserted.

    truncate_sheet : Boolean, optional
        Remove and recreate the destination sheet of the exported excel file.

    """
            
    extensions = ['shp', 'csv', 'parquet', 'xlsx']

    if 'geometry' in df.columns or ext == 'shp':
        rW.save_df_as_shp(df, path, name)
    elif ext == 'csv':
        rW.save_df_as_csv(df, path, name, usa)
    elif ext == 'parquet':
        rW.save_df_as_parquet(df, path, name)
    elif ext == 'xlsx':
        rW.save_df_as_excel(df, path, name, sheet_name, start_row, truncate_sheet)
    else: 
        raise Exception(f'ext parameter must be one of the following {extensions}')
    return None


def get_col(df, get_cols, from_df, key, key2=None):

    """
    Add columns to an input DataFrame, originated from another DataFrame
    based on a merge correspondence with in common column.

    Parameters
    ----------
    df : DataFrame
        Main input dataframe that will receive the merged columns.

    get_cols : List
        Columns from the other input dataframe you want to merge.

    from_df : DataFrame
        Secondary input dataframe that will be used to get the
        columns to be merged.

    key : String
        Column in main DataFrame used as key to merge the two DataFrames.

    key2 : String, optional
        Column in secondary DataFrame used as key to merge the two DataFrames.

    Returns
    -------
    ans : DataFrame
        Main input DataFrame with the merged columns.

    """

    if isinstance(key, str):
        key = [key]
    if isinstance(get_cols, str):
        get_cols = [get_cols]
    
    if key2 is not None:
        # If key2 is passed, then the keys are different in each df.
        if isinstance(key2, str):
            key2 = [key2]
        assert len(key) == len(key2), 'key and key2 must be of the same size'
        # Make dummy columns to avoid repetition and get only the desired cols
        for i, col in enumerate(key):
            df['temp'+str(i)] = df[col]
        for i, col in enumerate(key2):
            from_df['temp'+str(i)] = from_df[col]
        key = ['temp'+str(i) for i in range(len(key))]
    
    if isinstance(key, list) and isinstance(get_cols, list):
        cols = key + get_cols
    else:
        raise Exception('KEY and GET_COLS must be string or lists')
    
    # Perform the merge.
    # Using the method not the function since function fails when df is GeoPandas.
    # Assure types.
    from_df[key] = from_df[key].astype(df[key].dtypes)
    ans = df.merge(from_df[cols].drop_duplicates(key), on=key, how='left')
    
    if len(ans) != len(df):
        print(len(ans), len(df))
        raise Exception('Error performing get_col: ON not right')
    
    if key2 is not None:
        # Delete the temporary columns
        ans.drop(columns=key, inplace=True)
        from_df.drop(columns=key, inplace=True) 
    return ans


def get_mask_isin(df, key, base_data):

    """
    Get a filter mask to be used in a DataFrame based on
    single ou multiple filter.
    (Easier than the usual way)

    Parameters
    ----------
    df : DataFrame
        DataFrame that the user want to create a mask.

    key : String or List
        The column, or columns, that the filter will be applied.

    base_data : List and Tuple
        List filled with tuples that each one represents one filter.

    Returns
    -------
    mask : Series
        A Series of Boolean values that can be used as a mask for
        the input DataFrame.

    """

    if isinstance(key, list):
        new_col = '_'.join(key)
        my_list_series = []
        for k in key:
            my_list_series.append(list(df[k].values))
        df[new_col] = list(zip(*my_list_series))
        mask = df[new_col].isin(base_data)
        del df[new_col]
    else:    
        mask = df[key].isin(base_data)
    return mask


def remove_duplicate_safe(df, cols):

    """
    Remove DataFrame duplicates based on a column,
    or multiple columns, value.

    Parameters
    ----------
    df : DataFrame
        Dataframe that the user want to drop duplicates.

    cols : String or List
        Column or list of columns used as parameter for
        the drop_duplicates function.

    Returns
    -------
    df : DataFrame
        DataFrame without the duplicates in cols parameter.

    """

    num = len(df)
    df = df.drop_duplicates(cols)
    diff = num-len(df)
    if diff > 0:
        print(f'Duplicates removed from [{cols:}]{diff:,}{diff/num:,.2f}')
    return df


def flatten_hierarchical_col(col, sep='_'):

    """
    Turn multi-index column names into single index
    separated by '_'.

    Parameters
    ----------

    col : Multi-Index Column Title
        Multi-Index column that will be replaced for a single_index column.
    sep : String, optional
        Separator for the single_index column name.

    Returns
    -------
    new_col : Single_Index Column Title
        Multi_Index column title passed in the parameters, but flattened
        as no hierarchical title.

    """

    if not type(col) is tuple:
        new_col = col
    else:
        new_col = ''
        for leveli, level in enumerate(col):
            if not level == '':
                if not leveli == 0:
                    new_col += sep
                new_col += str(level)
    return new_col

# Faz o download automático de dados do IBGE/vários outros.

# def download():
    # return None



# -*- coding: utf-8 -*-
"""
Created on Thu Oct 22 16:18:33 2020

@author: bcalazans

Automatiza (na medidada do possível) o download de dados do FTP do Governo Brasileiro

Os dados podem ser usados em rotina futura para avaliação de dados socioeconomicos
"""

#%% imports
import pandas as pd
import zipfile as zf
import shutil
import glob, os
import sys
import py7zr
from datetime import datetime
from ftplib import FTP
import itertools

#%% FUNÇÕES AUXILIARES

def move(destination):
    ''' Move arquivos TXT ou CSV para mesmo diretório, deleta o resto'''
    all_files = []
    # para todos os diretórios filhos, lista abspath dos arquivos a mover
    for root, _dirs, files in itertools.islice(os.walk(destination), 1, None):
        for filename in files:
            if (filename[-3:] == 'txt') or (filename[-3:] == 'csv'):
                all_files.append(os.path.join(root, filename))
    # move todos os arquivos listados
    for filename in all_files:
        shutil.move(filename, destination)
    # deleta todos os diretórios
    for root, _dirs, files in itertools.islice(os.walk(destination), 1, None):
        shutil.rmtree(root)
    return
            
            
def unzip_All_Files(zipsDirectory, destinyDirectory):
    ''' Extrai todos os arquivos em uma pasta com o nome do zip de origem'''
    files = os.listdir(zipsDirectory)
    for file in files:
        if(file.endswith('.zip') or file.endswith('.7z')):
            print('Extraindo ' + file)
            unzip_file(os.path.join(zipsDirectory, file), 
                       os.path.join(destinyDirectory, file[:-4]))
    return        
        
    
def unzip_file(zipFileN, destinyDirectory):
    '''Unzipa arquivo no arquivo de destino. Avisa quando erro acontece'''
    try:
        with zf.ZipFile(zipFileN, 'r') as zip_ref:
            zip_ref.extractall(destinyDirectory)
# =============================================================================
#             # extracts all files AND fllatens hierarquy
#             for zip_info in zip_ref.infolist():
#                 if zip_info.filename[-1] == '/':
#                     continue
#                 zip_info.filename = os.path.basename(zip_info.filename)
#                 zip_ref.extract(zip_info, my_dir)            
# 
# =============================================================================

    except:
        try:
            #print("Ocorreu uma exceção ao extrair com a biblioteca Python ZipFile.")
            #print("Tentando extrair usando py7zr \n")
            archive = py7zr.SevenZipFile(zipFileN, mode='r')
            archive.extractall(path=destinyDirectory)
            archive.close()
        except:
            #print('Arquivo corrompido, em formato desconhecido ou com tipo de compressão não permitida, não foi possivel extrair o arquivo: ' + zipFileN + '\n')
            print('\t\t\tERRO! Extraia manualmente SOMENTE ARQUIVO DE TEXTO para a pasta\n\t\t\t' + destinyDirectory)

    return


def get_rais_zips(url, path, save_path):
    print("Executando RAIS")
    print()
    ftp =  FTP(url)
    ftp.login()
    #Endereço do servidor FTP. Não incluir o protocolo, senão não funciona!
    mainPath = path
    ftp.cwd(mainPath)
    #Listando as pastas para acompanhamento:
    print('Listando as pastas do FTP:')
    listFolders = ftp.nlst()
    print (listFolders)
    print('Entrando nas pastas e listando os arquivos:')
    rais_years = list(range(2003,2019)) #2003 a 2018
    for folder in listFolders:
        if(folder.isdigit() and int(folder) in rais_years):
            ftp.cwd(mainPath + '/' + folder)
            listSubFolders = ftp.nlst()
            for i in listSubFolders:
                if('estb' in i.lower() or 'estab' in i.lower() ):
#                     baixe
                    # Inicia o download:
                    filename = i 
                    print('Baixando o arquivo: '+filename)
                    with  open(os.path.join(save_path, filename), 'wb') as file:
                        ftp.retrbinary('RETR ' + filename, file.write)
                    os.rename(os.path.join(save_path, filename),
                              os.path.join(save_path, 'RAIS_'+dict_UF['UF']+folder+'.zip'))

        ftp.cwd(mainPath)
    #unzip_All_Files(save_path, save_path)
    
    
def get_censo_zips(url, path, save_path, **dict_UF):
    print("Executando CENSO")
    ftp =  FTP(url)
    ftp.login()

    #Endereço do servidor FTP. Não incluir o protocolo, senão não funciona!
    mainPath = path
    ftp.cwd(mainPath)

    #Listando as pastas para acompanhamento:
    print('Listando as pastas do FTP:')
    listFiles = ftp.nlst()
#     del(listFolders[9])
    print (listFiles)
    print('Entrando nas pastas e listando os arquivos:')
    for files in listFiles:
        if(files[:2] == dict_UF["UF"]):
            filename = files
            print('Baixando o arquivo: '+filename)
            with  open(os.path.join(save_path, filename), 'wb') as file:
                ftp.retrbinary('RETR ' + filename, file.write)
            os.rename(os.path.join(save_path, filename),
                      os.path.join(save_path, 'POPibge2010_'+dict_UF['UF']+'.zip'))
#         ftp.cwd(mainPath)
    #unzip_All_Files(save_path, save_path)
    #moveArchive(getExcelDirctory(save_path), local_path["censo_excel"])
    return
    
def get_cnefe_zips(url, path, save_path, **dict_UF):
    print("Executando CNEFE")
    ftp =  FTP(url)
    ftp.login()

    #Endereço do servidor FTP. Não incluir o protocolo, senão não funciona!
    mainPath = path
    ftp.cwd(mainPath)

    #Listando as pastas para acompanhamento:
    print('Listando as pastas do FTP:')
    listFolders = ftp.nlst()
    del(listFolders[9])
    print (listFolders)
    print('Entrando nas pastas e listando os arquivos:')
    for folder in listFolders:
        activePath = folder
        if (dict_UF is not None) and (activePath in dict_UF.values()):
            ftp.cwd(activePath)
            print('Pasta Atual'+activePath+' :')
            listFiles = ftp.nlst()
    #         print(listFiles)
            # Inicia o download:
            filename = listFiles[0]
            print('Baixando o arquivo: '+filename)
            with open(os.path.join(save_path, filename), 'wb') as file:
                ftp.retrbinary('RETR ' + filename, file.write)
            os.rename(os.path.join(save_path, filename),
                      os.path.join(save_path, 'CNEFE_'+dict_UF['UF']+'.zip'))
        ftp.cwd(mainPath)
    return
    #unzip_All_Files(save_path, save_path)
    
    
def get_pib_zips(url, path, save_path):
    print("Executando PIB")
    ftp =  FTP(url)
    ftp.login()

    #Endereço do servidor FTP. Não incluir o protocolo, senão não funciona!
    mainPath = path
    #print(mainPath)
    ftp.cwd(mainPath)
    #print('Pasta Atual'+mainPath+' :')
    listFiles = ftp.nlst()
    # Inicia o download:
    local_filename = os.path.join(save_path, "base_de_dados_2010_2017_txt.zip")
    filename = "base_de_dados_2010_2017_txt.zip"
    print('Baixando o arquivo: '+filename)
    with open(local_filename, 'wb') as file:
        ftp.retrbinary('RETR ' + filename, file.write)
    #unzip_All_Files(save_path, save_path)

#%% PARAMETERS
save_path = 'C:\\Users\\bcalazans\\Desktop\\getSocioeconomicsData\\data'

url ={
    "ftp_ibge": "ftp.ibge.gov.br",
    "ftp_microdados": "ftp.mtps.gov.br",
    "censo": "/Censos/Censo_Demografico_2010/Resultados_do_Universo/Agregados_por_Setores_Censitarios/",
    "cnefe": "/Censos/Censo_Demografico_2010/Cadastro_Nacional_de_Enderecos_Fins_Estatisticos/",
#     "pib": "/Pib_Municipios/2017/base/base_de_dados_2010_2017_txt.zip",
    "pib": "/Pib_Municipios/2017/base/",
    "rais": "/pdet/microdados/RAIS/",
    
    # TODO - trazer num vagas por setor censitario
    "ftp_inep" : "ftp.inep.gov.br",
    "microdados_edu_basica":"\\microdado\\microdados_educacao_basica_2018.zip\\microdados_ed_basica_2018\\DADOS\\TURMAS.zip\\TURMAS.CSV",
    "microdados_edu_superior":""
    
}

'''
   "vagas_educa":"\\microdado\\microdados_educacao_basica_2018.zip\\microdados_ed_basica_2018\\DADOS"
    ESCOLAS.zip//ESCOLAS.CSV 
        CO_DISTRITO, QT_FUNCIONARIOS (sum)
    MATRICULA_CO.zip//MATRICULA_CO.CSV  
        CO_DISTRITO, ID_ALUNO (count)
    etc... _NORDESTE _NORTE _SUL _SUDESTE
    TURMAS.zip/TURMAS.CSV
        CO_DISTRITO, QT_MATRICULAS (sum)
'''

write_mode = "w+"
dict_UF = {"UF": "RS", "code":43}
# https://atendimento.tecnospeed.com.br/hc/pt-br/articles/360021494734-Tabela-de-C%C3%B3digo-de-UF-do-IBGE

#%% begin   

# Rodar as funções necessárias para cada projeto
# CNEFE: dados de número de estabelecimento por setor censitário (SC)
# CENSO: dados de população do censo 2010 por IDADE e RENDA NOMINAL por SC
# PIB:   dados de PIB e PIB per capta por município
# RAIS:  dados de empregos formais por municipio 2013 a 2018

# dados municipais Brasil
get_pib_zips(url["ftp_ibge"], url["pib"], save_path)
get_rais_zips(url["ftp_microdados"], url["rais"], save_path) 

# dados por UF
dict_UF = {"UF": "RS", "code":43}
get_cnefe_zips(url["ftp_ibge"], url["cnefe"], save_path, **dict_UF)
get_censo_zips(url["ftp_ibge"], url["censo"], save_path, **dict_UF)

dict_UF = {"UF": "MG", "code":31}
get_cnefe_zips(url["ftp_ibge"], url["cnefe"], save_path, **dict_UF)
get_censo_zips(url["ftp_ibge"], url["censo"], save_path, **dict_UF)
 
dict_UF = {"UF": "PE", "code":26}
get_cnefe_zips(url["ftp_ibge"], url["cnefe"], save_path, **dict_UF)
get_censo_zips(url["ftp_ibge"], url["censo"], save_path, **dict_UF)
#%% UNZIP MANUAL
# Selecione todos os zips baixados e unzip manualmente
# é mais seguro do que rodar o código que falha em alguns PCs

# TODO -- extrair todos os arquivos para um mesmo diretório
# unzip_All_Files(save_path, save_path)
#%% move todos os arquivos de interesse pro dir raiz
move(save_path)
#%%




# -*- coding: utf-8 -*-
"""
REFERENCIAS:
    https://towardsdatascience.com/perform-regression-diagnostics-and-tackle-uncertainties-of-linear-models-1372a03b1f56
    https://zhiyzuo.github.io/Linear-Regression-Diagnostic-in-Python/
    https://boostedml.com/2018/08/testing-linear-regression-assumptions-the-kaggle-housing-price-dataset.html
    https://pythonfordatascienceorg.wordpress.com/linear-regression-python/
    https://blog.minitab.com/blog/adventures-in-statistics-2/understanding-t-tests-t-values-and-t-distributions#:~:text=A%20test%20statistic%20is%20a,data%20during%20a%20hypothesis%20test.&text=A%20t%2Dvalue%20of%200,of%20the%20t%2Dvalue%20increases.
    https://stats.stackexchange.com/questions/137498/how-to-interpret-the-direction-of-the-harvey-collier-test-and-rainbow-test-for-l

Created on Mon Aug 31 09:36:40 2020

@author: bcalazans

OBJETIVO GERAL
Criar uma rotina do Modelo 4-etapas rapidamente replicável para todas as cidades

SEQUENCIA DE PASSOS
## STEP 1 : Geração
    > Fazer levantamento de variáveis socioeconômicas explicativas
    > Fazer a matriz OD do TCol usando (SBE+GPS) (ou teremos uma ref externa)
    > Fazer um modelo de regressão

## STEP 2 : Distribuição

## STEP 3 : Divisão modal (+ projeção ?)

## STEP 4 : Alocação
    > TransCAD modelling 

------------------------
OBJETIVO
Definir um modelo de regreção pra ATRAÇÃO e outro pra PRODUÇÃO

Como:
    
    1. Determinação de variáveis explicativas que vão influenciar na geração 
        de viagens, ora na atração de viagens, ora na produção
        a. Esta etapa está associada a um passo preliminar que envolve 
        estabelecer/conhecer as principais motivações e comportamentos
        de viagens dentro do universo de estudo. Motivações genéricas 
        de viagens: por motivo trabalho; por motivo estudo.
        
        b. as variáveis explicativas selecionadas devem apresentar baixa 
        correlação entre si. Variáveis com forte relação de dependência 
        (alta correlação) riscam tendencionar os resultados da regressão 
        linear e  prejudicam a interpretação dos resultados 
        (análise da relação causa-efeito). 
             
    2. Determinação da relação de causa e efeito (modelo de regressão linear)
        a. Testar regressão para diferentes combinações de variáveis
        b. Testes possíveis para melhorar os resultados dos modelos de regressão:
            i. testar partindo ou não da origem
            ii. tirar outliers (tira de todos os BDs)
            
    3. Validação dos modelos de regressão:
        a. Verificar resultados dos testes estatísticos
            i. test-T,  p-value, F-score, R², etc
        b. Conferência dos coeficientes atribuídos a cada variável
            i. interpretação da equação de regressão final obtida
                > coeficientes negativos assosciados a alguma variável podem 
                ser contra-intuitivos;
                > constantes - coeficiente "b" - muito elevadas podem ser 
                contra-intuitivas em análises de projeção de demanda

INPUTS
    1. alpha = confiabilidade = 0,05 = 95% de confiabilidade
    2. Entrada de dados: associados a cada ZONA de tráfego
        (y) = atração OU produção = tiradas da OD SBE
        (x) = bancos de dados oficiais (IBGE, RAIS, etc)

OUTPUTS
Página HTML para cada regressão realizada, a página permite a análise crítica
do modelo gerado e decisão de qual equação utilizar
"""
import os
import warnings
import pandas as pd
import numpy as np
import seaborn as sns

import matplotlib.pyplot as plt
import statsmodels.api as sm
import statsmodels.stats.api as sms
import statsmodels.stats.stattools as st
import base64
from io import BytesIO
from itertools import combinations 
from tqdm import tqdm


def find_outliers(model, x, y, intercept, cov_type,plot=False):  
    '''
    Dado um modelo, re-faz regressões graduais a partir das 10 observações mais
    relevantes até incluir todas as observações.
    
    Analisa o resultado e retorna o modelo de maior R² dentro de um limite de 
    retirar apenas 10% das observações

    Parameters
    ----------
    model : statsmodels.regression.linear_model.RegressionResultsWrapper
        modelo inicial base de regressão
    x : pandas.core.frame.DataFrame
        variáveis explicativas
    y : pandas.core.series.Series
        variável dependente
    intercept : bool
        True or False para interceptar o eixo em zero
    cov_type : string
        parâmetro de fit_model()

    Returns
    -------
    model_out : statsmodels.regression.linear_model.RegressionResultsWrapper
        modelo da melhor regressão sem outliers.
    dfR : pandas.core.frame.DataFrame
        lista das regressões progressivas realizadas.
    Xoutlier : pandas.core.frame.DataFrame
        variáveis independentes, sem os outliers.
    Youtlier : pandas.core.series.Series
        variável dependente, sem os outliers.

    '''
    model_out = None # will return an empty list if there is "no" outlier
    # para o modelo encontrado, refaz a regressão um a um para estudar a evolução do R²
    df = x.merge(y,left_index=True, right_index=True)
    df['RESID'] = abs(model.resid)
    df = df.sort_values(by=["RESID"],ascending=True)
    
    # realiza regressões cada hora aumentando o número de observações
    dfR=pd.DataFrame(columns=['R2','NofObservations','Zones'], dtype=float)
    for i in range(10, df.shape[0]+1):
        zones = df.index[:i].values
        if intercept: model_out = sm.OLS(y.loc[zones], sm.add_constant(x.loc[zones])).fit(cov_type=cov_type)
        else:  model_out = sm.OLS(y.loc[zones], x.loc[zones]).fit(cov_type=cov_type)

        new_line={'R2':model_out.rsquared,
                 'NofObservations':len(zones),
                 'Zones':zones.tolist()}
        dfR = pd.concat([dfR, 
                         pd.DataFrame.from_records([new_line])],ignore_index=True)
        
    min_zn_out = len(zones)-int(0.1*len(zones)) # Qtd total de zonas menos os 10% de outliers
    xmax = len(zones)
    xmin = dfR['NofObservations'].min()
    zn_rmax = int(dfR.set_index('NofObservations').loc[min_zn_out:,'R2'].idxmax())
    rmax = dfR.set_index('NofObservations').loc[zn_rmax,'R2']

    # Para gerar o mapa   
    mask = dfR.NofObservations.isin(list(range(zn_rmax+1,xmax+1)))
    dfR['Bool']=np.where(mask,1,0)
    
    fightml = ''
#   maphtml = ''
    ymin = dfR['R2'].min()
    if plot:
    # TODO mudar esse código para linear_regression_make e ver se tem como enxugar ele
        lastick = [xmax]
        plt.style.use('ggplot')
        plt.figure(figsize=(26.6667,15))
        plt.gcf().subplots_adjust(right=2)
        dfR.plot(x='NofObservations',y='R2',color='darkblue', label='R²')
        plt.xticks(list(plt.xticks()[0])+lastick)
        plt.axis([xmin-2, xmax, ymin-0.01, 1.005])
        plt.axvline(min_zn_out,label='10% outliers',color='darkred')
        plt.axhline(rmax,linestyle='--',color='black')
        plt.ylabel('Rsquared',
                   fontdict={'weight':'bold',
                             'size':10,
                             'fontfamily':'serif'})
        plt.xlabel('Number of Observations',
                   fontdict={'weight':'bold',
                             'size':10,
                             'fontfamily':'serif'})
        plt.text(xmax+0.3,rmax,str('R² max = {0:.4f}'.format(rmax) + " ({} zonas)".format(zn_rmax)))
        plt.text(min_zn_out-1.7,ymin-0.02,str(min_zn_out),fontsize=12, color='darkred')
        plt.legend(facecolor='white')
        plt.title('Evolução do R² com o aumento do número de observações',
                  fontdict={'family': 'arial',
                            'color': 'black',
                            'weight': 'bold',
                            'size': 11})
        tmpfile = BytesIO()
        plt.savefig(tmpfile, format='png', bbox_inches = "tight")
        encoded = base64.b64encode(tmpfile.getvalue()).decode('utf-8')
        fightml = '''<img src=\'data:image/png;base64,{}\'>'''.format(encoded)

#chama a função de gerar o mapa       
#      mapa_zn_out(zns,dfR)
    
    # Escolhe melhor regressão retirando no máx 10% dos outliers
    Xoutlier = []
    Youtlier = []
    from pandas.core.common import flatten    
    zones = dfR[dfR.NofObservations>=min_zn_out].query('R2==R2.max()').Zones.values.tolist()
    zones = list(flatten(zones))
    if len(zones) < len(y):
        Xoutlier = x.loc[zones]
        Youtlier = y.loc[zones]
        if intercept: model_out = sm.OLS(Youtlier, sm.add_constant(Xoutlier)).fit(cov_type=cov_type)
        else:  model_out = sm.OLS(Youtlier,Xoutlier).fit(cov_type=cov_type) 
        
    return model_out, dfR, Xoutlier, Youtlier, fightml  

#def mapa_zn_out(shp_zonas, dfR)

def corr_mtx(df):
    fig, ax = plt.subplots()
    mask = np.zeros_like(df.corr(), dtype=np.bool)
    mask[np.triu_indices_from(mask)] = True
    sns.heatmap(df.corr(), mask = mask, annot=True,vmin=-1, vmax=1, center=0, cmap='RdBu')
    return
         

def linear_harvey_collier(model):
    ''' Test fot linearity unsing Harvey-Collier teory
    Original function has a bug
    sms.linear_harvey_collier(model)
    https://github.com/statsmodels/statsmodels/pull/6727
    
    Call this one instead
    '''   
    skip = len(model.params)  # bug in linear_harvey_collier
    rr = sms.recursive_olsresiduals(model, skip=skip, alpha=0.95)
    #try: rr = sms.recursive_olsresiduals(model, skip=skip, alpha=0.95)
    #except: rr = sms.recursive_olsresiduals(model, skip=skip, alpha=0.95)
    #dfDirty = dfClean+0.00001*np.random.rand(n, 2)
    return st.stats.ttest_1samp(rr[3][skip:], 0)


def test_conditions(model):
    ''' Testa várias condições estatísticas que norteiam a validação ou não 
    de um modelo de regressãol linear. O ideal é investigar no HTML resumo
        condition0 - verifica se há variáveis 100% correlacionadas
        condition1 - teste p-val < 0.05 e t-val> 2
        condition2 - harvey-colier (linearidade) > 0.05
        condition3 - DurbinWatson entre 1 e 3
    
    retorna VERDADEIRO apenas se passar em todas as condições acima
    '''
    condition1 = condition2 = condition3 = False
    
    # verifica se há variáveis 100% correlacioandas (ou colunas duplicadas)
    if len(model.params) == 1: condition0=True,
    else:
        corr = np.corrcoef(model.model.exog.T)
        mask = np.eye(len(model.model.exog_names), dtype = bool)
        non_diagonal_elements = corr[~mask]
        # remove potenciais NaN que acontece qndo tem CONSTANTE
        non_diagonal_elements = non_diagonal_elements[~np.isnan(non_diagonal_elements)]
        # se fora da diagnonal contem corr=1 então não deveria gerarl modelo!
        condition0 = (non_diagonal_elements < 0.9999).all()

    if condition0:
        # testa p-value e test T
        values = pd.DataFrame([abs(model.pvalues), abs(model.tvalues)], 
                              index=['p-val', 't-val']).T
        condition1 = ((values['p-val'] < 0.05) & 
                    (values['t-val'] > 2)).all()
        # testa harvey-colier > 5%
        #condition2 = sms.linear_harvey_collier(model).pvalue > 0.05
        try:condition2 = linear_harvey_collier(model).pvalue > 0.05
        except: 
            #print('There still a bug in HARVEY for singular matrix...')
            #print('\t', model.model.exog_names)
            #condtion2 = False
            raise Exception(model.model.exog_names)
        # Durbin-Watson aceitável para análise
        condition3 = ((1 < st.durbin_watson(model.resid, axis=0)) &
                      (3 > st.durbin_watson(model.resid, axis=0)))
   
    return condition0 & condition1 & condition2 & condition3


def get_combinations(cols):
    '''Get all combinations of the X variabels'''
    comb = []
    for i in tqdm(range(len(cols)), desc='Getting combinations for all variables'): 
        comb.extend(list(combinations(cols, i)) )
    del comb[0]
    print('{:,} possibilites of regression to do!'.format(len(comb)))
    return comb

    
def get_correlated_pairs(df, Xcols, LIMIT=0.5):
    ''' return a list of list of correlated enodg variables
    LIMIT: consider correlated with coef > LIMIT
    '''
    print('Get correlated pairs...')
    # get correlation values
    corr = df[Xcols].corr()
    # filter diagonal
    mask_diagonal = np.eye(len(Xcols), dtype = bool)
    # filter ultra correlated values
    mask_correlated = corr > LIMIT
    # filter upper triangle
    mask_upperTriangle = np.zeros_like(corr, dtype=np.bool)
    mask_upperTriangle[np.triu_indices_from(mask_upperTriangle)] = True
    
    # get True/False for ultra correlated
    corr = (~mask_diagonal) & mask_correlated & mask_upperTriangle

    # get names that are REALLY correlated
    corr = corr.reset_index().melt(id_vars='index').query('value == True')
    corr_pairs = corr[['index','variable']].values.tolist()
    return corr_pairs
  
    
def get_mean_and_std_from_kde(residuals):
    ''' Get meand and std values from the kde of the residuals
    não consegui fazer usando fórmula, então pra andar rápido usei o mesmo código
    que está em _plot'''
    import scipy as sp
    fig, (aq, ax) = plt.subplots(1,2, figsize=(16,5))
    _, (__, ___, r) = sp.stats.probplot(residuals, plot=aq, fit=True)
    aq.spines['top'].set_visible(False)
    aq.spines['right'].set_visible(False)

    # Create distplot in matplolib (there is no kde func in plotly)
    #fig, ax = plt.subplots()
    ax=residuals.plot.kde(ax=ax, title='Histograma dos residuais')
    mean_val = np.mean(ax.get_children()[0]._x)
    std_val = np.std(ax.get_children()[0]._x)
    plt.close(fig)
    return mean_val, std_val


def save_model(model, df_regs, intercept, condition):
    ''' Salva no DataFrame algumas informações sobre uma regressão gerada '''
    mean, std = get_mean_and_std_from_kde(model.resid)
    condition = 'SIM' if condition else 'NOP'
    df_regs = df_regs.append({'x':list(model.params.index), 
                              'Rsquared':model.rsquared,
                              'intercept':intercept,
                              'covType': model.cov_type,
                              'ZONAS': model.fittedvalues.index.values.tolist(),
                              'PASS_TESTS': condition,
                              'NUM_COEF_NEGATIVO':(model.params < 0).sum(),
                              'HARVEY_COLLIER':linear_harvey_collier(model).pvalue,
                              'DURBIN_WATSON':st.durbin_watson(model.resid, axis=0),
                              'MEAN_KDE':mean,
                              'STD_KDE':std}, 
                             ignore_index=True)
    
    return df_regs


def drop_combinations(myComb, corr_pairs, Xcols):
    '''drop combinations that we do not want to do a regression since contains 
    a pair of correlated variables'''
    
    myComb = pd.DataFrame({'combinations':myComb})
    myComb = myComb[myComb.combinations.apply(len)>1] # tira um prq está como tuple
    
    # build matrix that each column is the name of endog variable and 
    # each row sets to 1 if the column belogns to that combination arrange
    #temp = myComb.sample(10)
    temp = myComb.combinations.explode().reset_index().groupby(['index','combinations']).size().unstack()
    myComb = myComb.merge(temp, left_index=True, right_index=True)
    # if combination (row) contains a pair of correlated variables, dropit
    for pair in tqdm(corr_pairs, position=0, leave=True, desc="Droping combinations with correlated values..."):
        msk1 = myComb[pair[0]] == 1
        msk2 = myComb[pair[1]] == 1
        myComb = myComb[~(msk1 & msk2)]
    
    # add possibilidade de valores únicos
    myComb = myComb.combinations.tolist() + Xcols 
    print('{:,} non-correlated possibilities to do!'.format(len(myComb)))    
    return myComb



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  5 16:58:55 2018

@author: bcalazans

Miscellaneous file with functions that can/may be used by everyone
"""
import os
import numbers
import pandas as pd
import numpy as np
import myAux.readWrite as rw

from openpyxl import load_workbook
from inspect import currentframe, stack
from datetime import datetime

glob_count = 0
def test():
    print("hi from misc!")    
    return 1


# %% Basic functions
class CColor:
   """
   \033[       Escape code, this is always the same
   1           Style, (0 or 1 = normal, 3 = italic, 4 = underline)
   32          Text colour, 32 for white.
   40m         Background colour, 40 is for black.
   """
   BLUE = '\033[0;34;m'
   GREEN = '\033[0;32;m'
   RED = '\033[0;31;m'     

   ITALIC = '\033[3;30;m'         
   UNDERLINE = '\033[4;30;m'    
   ENDC = '\033[0;30;m'

   LAND = '#F9F6D8'
   SEA = '#DCF0FA'
   BORDER = '#696969'
   STREET = '#C0C0C0'
   

def log(list2print, option=None):
    """
    Prints an information on the prompt based on the list of items
    :param list2print: list of items to print
    :param option: configure stringformat adding style to it
    :return: nothing

    ------ Examples of use
    from numpy.random import randint
    df = pd.DataFrame(randint(0,3,size=(10, 4)), columns=list('ABCD'))
    mask = df.A >= 1
    df['C'] = df['D']/0.5
    log(['Testring my log', 123456789, 9.876])
    log(['Column A >= 1', mask])
    log(['Sum colum B', df.B.sum()])
    log(['Sum colum C', df.C.sum()])
    log(['Dropping A >= 1', mask], option='w')
    log([df])
    """
    # check passed parameters
    options = ['w', 'warning', 'u', 'underline', 'i', 'italic',
                'blue', 'green', 'per', None]
    assert option in options, 'Param' + option + 'unknown.\n ' \
                              'Options are: ' + options

    # build string to print
    my_str = ''
    msg_error = 'log function not structured to print that'
    for pos, val in enumerate(list2print[:]):
        if isinstance(val, pd.core.frame.DataFrame):
            #header = list(val.columns)
            #header = '\t'.join(str(e) for e in header)
            #print(CColor.UNDERLINE + header + CColor.ENDC)
            #print(val.to_csv(sep='\t', index=False, header=None))
            print(val)
            list2print.remove(val)
        elif isinstance(val, str): my_str = my_str + '{:<60}  '
        elif isinstance(val, list): 
            my_str = my_str + '{:<60}  '
            list2print[pos] = ', '.join(val)
        elif isinstance(val, numbers.Integral): my_str = my_str + '{:>12,}  '
        elif isinstance(val, numbers.Real):
            if option == 'per':
                my_str = my_str + '{:>10.2%}  '
            else:
                my_str = my_str + '{:>10,.2}  '
        elif isinstance(val, pd.core.series.Series):
            # if is a boolean mask, will compute the percentual
            if pd.api.types.is_bool_dtype(val):
                if len(val) > 0 :tot = len(val)
                else: tot = 1
                list2print[pos] = val.sum()
                list2print = list2print[:pos+1] \
                             + [(val.sum()/tot)] \
                             + list2print[pos+1:]
                my_str = my_str + '{:>12,}  {:>10.2%}  '
            else:
                raise Exception(msg_error)
        else:
            raise Exception(msg_error)

    # add style to string
    if option in ['w', 'warning']: my_str = CColor.RED + my_str 
    elif option in ['u', 'underline']: my_str = CColor.UNDERLINE + my_str 
    elif option in ['i', 'italic']: my_str = CColor.ITALIC + my_str 
    elif option in ['blue']: my_str = CColor.BLUE + my_str 
    elif option in ['green']: my_str = CColor.GREEN + my_str
    elif option == 'per': pass
    elif option is None: pass
    else:
        raise Exception('Forgot to add option here!')

    # print and return        
    my_str = my_str + CColor.ENDC   
    print(my_str.format(*list2print))
    # Call me to log on file
    #print_on_file(my_str.format(*list2print))
    return


def print_on_file(text):
    name_file = str(datetime.today().date()) + '.log'
    with open(name_file, "a") as f:
        f.write(text)
    return


def my_print(str2print):
    global glob_count
    glob_count = 1 # do smth to count how many functions was called out
    str2print = '\t'*glob_count + str2print
    print(str2print)
    return



def generate_od_pairs(list_of_ids):
    ''' Generates all combinations and return a data frame'''
    from itertools import product
    pairs = list(product(list_of_ids, repeat=2))
    return pd.DataFrame(pairs, columns=['ORG','DST'])



# -*- coding: utf-8 -*-


# -------------------------------------------------------------------------------------------------------------------- #
# Name:       CalcDistribution
# Purpose:    Utilities for various calculations of different types of trip distribution models.
#             a) calc_fratar : Calculates a Fratar/IPF on a seed matrix given row and column (P and A) totals
#             b) calc_singly_constrained : Calculates a singly constrained trip distribution for given P/A vectors and a
#                friction factor matrix
#             c) calc_doubly_constrained : Calculates a doubly constrained trip distribution for given P/A vectors and a
#                friction factor matrix (P and A should be balanced before usage, if not then A is scaled to P)
#             d) calc_multi_fratar : Applies fratar model to given set of trip matrices with multiple target production
#                vectors and one attr_action vector
#             e) calc_multi_distribute : Applies gravity model to a given set of frication matrices with multiple
#                production vectors and one target attr_action vector
#
#              **All input vectors are expected to be numpy arrays
#
# Author:      Chetan Joshi, Portland OR
# Dependencies:numpy [www.numpy.org]
# Created:     5/14/2015
#
# Copyright:   (c) Chetan Joshi 2015
# Licence:     Permission is hereby granted, free of charge, to any person obtaining a copy
#              of this software and associated documentation files (the "Software"), to deal
#              in the Software without restriction, including without limitation the rights
#              to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#              copies of the Software, and to permit persons to whom the Software is
#              furnished to do so, subject to the following conditions:
#
#              The above copyright notice and this permission notice shall be included in all
#              copies or substantial portions of the Software.
#
#              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#              IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#              FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#              AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#              LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#              OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#              SOFTWARE.
# -------------------------------------------------------------------------------------------------------------------- #
import numpy


def calc_fratar(prod_a, attr_a, trips_1, max_iter=10, print_balance=False):
    """
    Calculates fratar trip distribution
       prod_a = Production target as array
       attr_a = attr_action target as array
       trips_1 = Seed trip table for fratar
       max_iter (optional) = maximum iterations, default is 10
       Returns fratared trip table
    """
    # print('Checking production, attr_action balancing:')
    sum_p = sum(prod_a)
    sum_a = sum(attr_a)
    # print('Production: ', sum_p)
    # print('attr_action: ', sum_a)
    if sum_p != sum_a:
        if print_balance:
            print('Productions and attr_actions do not balance, attr_actions will be scaled to productions!')
        attr_a = attr_a*(sum_p/sum_a)
    else:
        if print_balance:
            print('Production, attr_action balancing OK.')
    # Run 2D balancing --->
    for balIter in range(0, max_iter):
        computed_productions = trips_1.sum(1)
        computed_productions[computed_productions == 0] = 1
        orig_fac = (prod_a/computed_productions)
        trips_1 = trips_1*orig_fac[:, numpy.newaxis]

        computedattr_actions = trips_1.sum(0)
        computedattr_actions[computedattr_actions == 0] = 1
        dest_fac = (attr_a/computedattr_actions)
        trips_1 = trips_1*dest_fac
    return trips_1


def calc_singly_constrained(prod_a, attr_a, f):
    """
    Calculates singly constrained trip distribution for a given friction factor matrix
    prod_a = Production array
    attr_a = attr_action array
    f = Friction factor matrix
    Resutrns trip table
    """
    sum_ajfij = (attr_a*f).sum(1)
    sum_ajfij[sum_ajfij == 0] = 0.0001
    return prod_a*(attr_a*f).transpose()/sum_ajfij


def calc_doubly_constrained(prod_a, attr_a, f, max_iter=10, verbose=False):
    """
    Calculates doubly constrained trip distribution for a given friction factor matrix
    prod_a = Production array
    attr_a = attr_action array
    f = Friction factor matrix
    max_iter (optional) = maximum iterations, default is 10
    Returns trip table
    """
    worse = f.min()
    f = f / worse

    trips_1 = numpy.zeros((len(prod_a), len(prod_a)))
    if verbose:
        print('Checking production, attr_action balancing:')
    sum_p = sum(prod_a)
    sum_a = sum(attr_a)
    if verbose:
        print('Production: ', sum_p)
        print('attr_action: ', sum_a)
    if sum_p != sum_a:
        if verbose:
            print('Productions and attr_actions do not balance, attr_actions will be scaled to productions!')
        attr_a = attr_a*(sum_p/sum_a)
        attr_t = attr_a.copy()
        prod_t = prod_a.copy()
    else:
        if verbose:
            print('Production, attr_action balancing OK.')
        attr_t = attr_a.copy()
        prod_t = prod_a.copy()

    for balIter in range(0, max_iter):
        for i in range(0, len(prod_a)):
            trips_1[i, :] = prod_a[i]*attr_a*f[i, :]/max(0.000001, sum(attr_a * f[i, :]))

        # Run 2D balancing --->
        computedattr_actions = trips_1.sum(0)
        computedattr_actions[computedattr_actions == 0] = 1
        attr_a = attr_a*(attr_t/computedattr_actions)

        computed_productions = trips_1.sum(1)
        computed_productions[computed_productions == 0] = 1
        prod_a = prod_a*(prod_t/computed_productions)

    for i in range(0, len(prod_a)):
        trips_1[i, :] = prod_a[i]*attr_a*f[i, :]/max(0.000001, sum(attr_a * f[i, :]))

    return trips_1


def calc_multi_fratar(prods, attr, trip_matrices, max_iter=10):
    """
    Applies fratar model to given set of trip matrices with target productions and one attr_action vector
    prods = Array of Productions (n production segments)
    attr_att = Array of attr_action ( 1 attr_action segment)
    trip_matrices = N-Dim array of seed trip matrices corresponding to prod_atts
     --> (num_trip_mats, num_zones, num_zones)
    max_iter = Maximum number of iterations
    version 1.0
    """
    num_zones = len(attr)
    num_trip_mats = len(trip_matrices)
    trip_matrices = numpy.zeros((num_trip_mats, num_zones, num_zones))

    # Run 2D balancing --->
    for Iter in range(0, max_iter):
        # computedattr_actions = numpy.ones(num_zones)
        computedattr_actions = trip_matrices.sum(1).sum(0)
        computedattr_actions[computedattr_actions == 0] = 1
        dest_fac = attr/computedattr_actions

        for k in range(0, len(num_trip_mats)):
            trip_matrices[k] = trip_matrices[k]*dest_fac
            computed_productions = trip_matrices[k].sum(1)
            computed_productions[computed_productions == 0] = 1
            orig_fac = prods[:, k]/computed_productions  # P[i, k1, k2, k3]...
            trip_matrices[k] = trip_matrices[k]*orig_fac[:, numpy.newaxis]

    return trip_matrices


def calc_multi_distribute(prods, attr, fric_matrices, max_iter=10):
    """
    prods = List of Production attributes
    attr  = attr_action attribute
    fric_matrices = N-Dim array of friction matrices corresponding to prod_atts --> 
    (numFrictionMats, num_zones, num_zones)
    max_iter (optional) = Maximum number of balancing iterations, default is 10
    Returns N-Dim array of trip matrices corresponding to each production segment
    """
    num_zones = len(attr)
    trip_matrices = numpy.zeros(fric_matrices.shape)
    num_fric_mats = len(fric_matrices)

    prod_op = prods.copy()
    attr_op = attr.copy()

    for Iter in range(0, max_iter):
        # Distribution --->
        for k in range(0, num_fric_mats):
            for i in range(0, num_zones):
                if prod_op[i, k] > 0:
                    trip_matrices[k, i, :] = prod_op[i, k] * attr_op * fric_matrices[k, i, :] / \
                                            max(0.000001, sum(attr_op * fric_matrices[k, i, :]))
        # Balancing --->
        computedattr_actions = trip_matrices.sum(1).sum(0)
        computedattr_actions[computedattr_actions == 0] = 1
        attr_op = attr_op*(attr/computedattr_actions)
        for k in range(0, len(fric_matrices)):
            computed_productions = trip_matrices[k].sum(1)
            computed_productions[computed_productions == 0] = 1
            orig_fac = prods[:, k]/computed_productions
            prod_op[:, k] = orig_fac*prod_op[:, k]
    # Final Distribution --->
    for k in range(0, num_fric_mats):
        for i in range(0, num_zones):
            if prod_op[i, k] > 0:
                trip_matrices[k, i, :] = prod_op[i, k] * attr_op * fric_matrices[k, i, :] / \
                                        max(0.000001, sum(attr_op * fric_matrices[k, i, :]))

    return trip_matrices



from setuptools import setup


setup(
  
    name='systool',
    version='0.05.00',
    license='MIT',
    author="Bruna Calazans (bruna-calazans) & Pedro Chiachio (pcardoso-sk)",
    author_email='bcalazans@systra.com',
    package_dir={'': 'main'},
    include_package_data=True,
    package_data={'utils': ['charts.py', 'flatt_geom.py', 'get_ftp_data.py', 'get_socioEconomic_data.py', 
                            'maps.py', 'readWrite.py', 'report.py', 'linear_regression_make.py',
                            'linear_regression_plot.py', 'miscFunctions']},
    url='https://github.com/bruna-calazans/systools',
    keywords=['systool', 'engenharia', 'transportes', 'systra', 'engenharia de transportes'],
    install_requires=['pytest',
                      'pandas',
                      'numpy',
                      'matplotlib',
                      'matplotlib_scalebar',
                      'geopandas',
                      'openpyxl',
                      'seaborn',
                      'statsmodels',
                      'tqdm',
                      'scipy',
                      'plotly',
                      'shapely'],
)
